# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Iterator, Mapping, Sequence, TypeVar

import numpy as np
from loguru import logger

from qilisdk.core.exceptions import EvaluationError, InvalidBoundsError, NotSupportedOperation, OutOfBoundsException
from qilisdk.yaml import yaml

if TYPE_CHECKING:
    from ruamel.yaml.nodes import ScalarNode
    from ruamel.yaml.representer import RoundTripRepresenter

Number = int | float | complex
RealNumber = int | float
GenericVar = TypeVar("GenericVar", bound="Variable")
CONST_KEY = "_const_"
MAX_INT = np.iinfo(np.int64).max
MIN_INT = np.iinfo(np.int64).min
LARGE_BOUND = 100


def LT(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Less Than' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs < rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.LT)


LessThan = LT


def LEQ(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Less Than or equal to' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs <= rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.LEQ)


LessThanOrEqual = LEQ


def EQ(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Equal to' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs == rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.EQ)


Equal = EQ


def NEQ(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Not Equal to' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs != rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.NEQ)


NotEqual = NEQ


def GT(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Greater Than' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs > rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.GT)


GreaterThan = GT


def GEQ(lhs: RealNumber | BaseVariable | Term, rhs: RealNumber | BaseVariable | Term) -> ComparisonTerm:
    """'Greater Than or equal to' mathematical operation

    Args:
        lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
        rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.

    Returns:
        ComparisonTerm: a comparison term with the structure lhs >= rhs.
    """
    return ComparisonTerm(lhs=lhs, rhs=rhs, operation=ComparisonOperation.GEQ)


GreaterThanOrEqual = GEQ


def _extract_number(label: str) -> int:
    """Extracts the number from the variable's label.

    Args:
        label (str): variable label that follows the format <label>(<number>).

    Returns:
        int: the number in the label.
    """
    pattern = re.compile(r"\((\d+)\)$")
    matches = pattern.search(label)
    if matches is not None:
        return int(matches.group(1))
    return 0


def _float_if_real(value: Number) -> Number:
    if isinstance(value, RealNumber):
        return value
    if isinstance(value, complex) and value.imag == 0:
        return value.real
    return value


def _assert_real(value: Number) -> RealNumber:
    _value = _float_if_real(value)
    if isinstance(_value, RealNumber):
        return _value
    raise ValueError(f"Only Real values are allowed but {_value} was provided.")


@yaml.register_class
class Domain(str, Enum):
    INTEGER = "Integer Domain"
    POSITIVE_INTEGER = "Positive Integer Domain"
    REAL = "Real Domain"
    BINARY = "Binary Domain"
    SPIN = "Spin Domain"

    def check_value(self, value: Number) -> bool:
        """checks if the provided value is valid for a given domain

        Args:
            value (int | float): the value to be evaluated.

        Returns:
            bool: True if the value provided is valid, False otherwise.
        """
        if self == Domain.BINARY:
            return isinstance(value, Number) and value in {0, 1}
        if self == Domain.SPIN:
            return isinstance(value, Number) and value in {-1, 1}
        if self == Domain.REAL:
            return isinstance(value, (int, float))
        if self == Domain.INTEGER:
            return isinstance(value, int)
        if self == Domain.POSITIVE_INTEGER:
            return isinstance(value, int) and value >= 0
        return False

    def min(self) -> float:
        """
        Returns:
            float: the minimum value allowed of a given domain.
        """
        if self in {Domain.BINARY, Domain.POSITIVE_INTEGER}:
            return 0
        if self == Domain.SPIN:
            return -1
        if self == Domain.INTEGER:
            return MIN_INT
        return -1e30

    def max(self) -> float:
        """
        Returns:
            float: the maximum value allowed for a given domain.
        """
        if self in {Domain.BINARY, Domain.SPIN}:
            return 1
        if self in {Domain.POSITIVE_INTEGER, Domain.INTEGER}:
            return MAX_INT
        return 1e30

    @classmethod
    def to_yaml(cls, representer: RoundTripRepresenter, node: Domain) -> ScalarNode:
        """
        Method to be called automatically during YAML serialization.

        Returns:
            ScalarNode: The YAML scalar node representing the Domain.
        """
        return representer.represent_scalar("!Domain", f"{node.value}")

    @classmethod
    def from_yaml(cls, _, node: ScalarNode) -> Domain:
        """
        Method to be called automatically during YAML deserialization.

        Returns:
            Domain: The Domain instance created from the YAML node value.
        """
        return cls(node.value)


@yaml.register_class
class Operation(str, Enum):
    MUL = "*"
    ADD = "+"
    DIV = "/"
    SUB = "-"

    @classmethod
    def to_yaml(cls, representer: RoundTripRepresenter, node: Operation) -> ScalarNode:
        """
        Method to be called automatically during YAML serialization.

        Returns:
            ScalarNode: The YAML scalar node representing the Operation.
        """
        return representer.represent_scalar("!Operation", f"{node.value}")

    @classmethod
    def from_yaml(cls, _, node: ScalarNode) -> Operation:
        """
        Method to be called automatically during YAML deserialization.

        Returns:
            Operation: The Operation instance created from the YAML node value.
        """
        return cls(node.value)


@yaml.register_class
class ComparisonOperation(str, Enum):
    LT = "<"
    LEQ = "<="
    EQ = "=="
    NEQ = "!="
    GT = ">"
    GEQ = ">="

    @classmethod
    def to_yaml(cls, representer: RoundTripRepresenter, node: ComparisonOperation) -> ScalarNode:
        """
        Method to be called automatically during YAML serialization.

        Returns:
            ScalarNode: The YAML scalar node representing the ComparisonOperation.
        """
        return representer.represent_scalar("!ComparisonOperation", f"{node.value}")

    @classmethod
    def from_yaml(cls, _, node: ScalarNode) -> ComparisonOperation:
        """
        Method to be called automatically during YAML deserialization.

        Returns:
            ComparisonOperation: The ComparisonOperation instance created from the YAML node value.
        """
        return cls(node.value)


@yaml.register_class
class Encoding(ABC):
    """Represents an abstract variable encoding class.

    The Encoding is defined on the variable bases, and it defines how the continuous variables are encoded into binary
    variables.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Encoding's name

        Returns:
            str: The name of the encoding.
        """

    @staticmethod
    @abstractmethod
    def encode(var: Variable, precision: float = 1e-2) -> Term:
        """Given a continuous variable return a Term that only consists of
            binary variables that represent the continuous variable in the given encoding.

        Args:
            var (ContinuousVar): The continuous variable to be encoded
            precision (int): the precision to be considered for real variables (Only applies if
                                the variable domain is Domain.Real)

        Returns:
            Term: a term that only contains binary variables
        """

    @staticmethod
    @abstractmethod
    def encoding_constraint(var: Variable, precision: float = 1e-2) -> ComparisonTerm:
        """Given a continuous variable return a Constraint Term that ensures that the encoding is respected.

        Args:
            var (ContinuousVar): The continuous variable to be encoded
            precision (float): the precision to be considered for real variables (Only applies if
                                the variable domain is Domain.Real)

        Returns:
            Constraint Term: a constraint term that ensures the encoding is respected.
        """

    @staticmethod
    @abstractmethod
    def evaluate(var: Variable, value: list[int] | int, precision: float = 1e-2) -> float:
        """Given a binary string, evaluate the value of the continuous variable in the given encoding.

        Args:
            var (ContinuousVar): the variable to be evaluated
            value (list[int] | int): a list of binary values or an integer value.
            precision (float): the precision to be considered for real variables (Only applies if
                                the variable domain is Domain.Real)

        Returns:
            float: the value of the continuous variable given the specified binary values.
        """

    @staticmethod
    @abstractmethod
    def num_binary_equivalent(var: "Variable", precision: float = 1e-2) -> int:
        """Give a continuous variable return the number of binary variables needed to encode it.

        Args:
            var (ContinuousVar): the continuous variable.
            precision (float): the precision to be considered for real variables (Only applies if
                                the variable domain is Domain.Real)

        Returns:
            int: the number of binary variables needed to encode it.
        """

    @staticmethod
    @abstractmethod
    def check_valid(value: list[int] | int) -> tuple[bool, int]:
        """checks if the binary list sample is a valid sample in this encoding.

        Args:
            value (list[int] | int):  a list of binary values or an integer value.

        Returns:
            tuple[bool, int]: the boolean is True if the sample is a valid encoding,
                                while the int is the error in the encoding.
        """


@yaml.register_class
class Bitwise(Encoding):
    """Represents a Bitwise variable encoding class."""

    name = "Bitwise"

    @staticmethod
    def _bitwise_encode(x: int, N: int) -> list[int]:
        """encode the integer x in Bitwise encoding.

        Args:
            x (int): the integer to be encoded.
            N (int): the number of bits to encode x.

        Returns:
            list[int]: a binary list representing the Bitwise encoding of the integer x.
        """
        return list(reversed([int(b) for b in format(x, f"0{N}b")]))

    @staticmethod
    def encode(var: Variable, precision: float = 1e-2) -> Term:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        abs_bound = np.abs(bounds[1] - bounds[0])
        n_binary = int(np.floor(np.log2(abs_bound if abs_bound != 0 else 1)))
        binary_vars = [BinaryVariable(var.label + f"({i})") for i in range(n_binary + 1)]

        term = sum(2**i * binary_vars[i] for i in range(n_binary))
        term += (np.abs(bounds[1] - bounds[0]) + 1 - 2**n_binary) * binary_vars[-1]
        term += bounds[0]
        return term * var.precision if var.domain is Domain.REAL else term

    @staticmethod
    def evaluate(var: Variable, value: list[int] | int, precision: float = 1e-2) -> float:
        term = Bitwise.encode(var, precision)
        binary_var = sorted(
            term.variables(),
            key=lambda x: _extract_number(x.label),
        )

        binary_list = Bitwise._bitwise_encode(value, len(binary_var)) if isinstance(value, Number) else value

        if not Bitwise.check_valid(binary_list)[0]:
            raise ValueError(
                f"invalid binary string {binary_list} with the Bitwise encoding."
            )  # can not be reached in the case of Bitwise encoding.

        if len(binary_list) < len(binary_var):
            for _ in range(len(binary_var) - len(binary_list)):
                binary_list.append(0)
        elif len(binary_list) != len(binary_var):
            raise ValueError(f"expected {len(binary_var)} variables but received {len(binary_list)}")

        binary_dict: dict[BaseVariable, list[int]] = {binary_var[i]: [binary_list[i]] for i in range(len(binary_list))}

        _out = term.evaluate(binary_dict)

        if isinstance(_out, RealNumber):
            out = float(_out)
        elif isinstance(_out, complex) and _out.imag == 0:
            out = float(_out.real)
        else:
            raise ValueError(f"Evaluation answer ({_out}) is outside the variable domain ({var.domain}).")

        out = int(out) if var.domain in {Domain.INTEGER, Domain.POSITIVE_INTEGER} else out

        if not var.domain.check_value(out):
            raise ValueError(
                f"The value {out} violates the domain {var.domain.__class__.__name__} of the variable {var}"
            )  # not sure this line can be reached.
        return out

    @staticmethod
    def encoding_constraint(var: Variable, precision: float = 1e-2) -> ComparisonTerm:
        raise NotImplementedError("Bitwise encoding constraints are not supported at the moment")

    @staticmethod
    def num_binary_equivalent(var: "Variable", precision: float = 1e-2) -> int:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.floor(np.log2(np.abs(bounds[1] - bounds[0]))))

        return n_binary + 1

    @staticmethod
    def check_valid(value: list[int] | int) -> tuple[bool, int]:
        return True, 0


@yaml.register_class
class OneHot(Encoding):
    """Represents a One-Hot variable encoding class."""

    name = "One-Hot"

    @staticmethod
    def _one_hot_encode(x: int, N: int) -> list[int]:
        """One-hot encode integer x in range [0, N-1].

        Args:
            x (int): the value to be encoded
            N (int): the number of bits to encode x.

        Raises:
            ValueError: if x is larger than N or less than 0

        Returns:
            list[int]: a binary list representing the one hot encoding of the integer x.
        """
        if not (0 <= x < N):
            raise ValueError(f"the input value ({x}) must be in range [0, {N - 1}]")
        return [1 if i == x else 0 for i in range(N)]

    @staticmethod
    def _find_zero(var: Variable) -> int:
        binary_var = var.bin_vars
        term = var.term
        for i in range(var.num_binary_equivalent()):
            if binary_var[i] not in term:
                return i
        return 0

    @staticmethod
    def encode(var: Variable, precision: float = 1e-2) -> Term:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0])) + 1

        binary_vars = [BinaryVariable(var.label + f"({i})") for i in range(n_binary)]

        term = Term([(bounds[0] + i) * binary_vars[i] for i in range(n_binary)], Operation.ADD)

        return term * var.precision if var.domain is Domain.REAL else term

    @staticmethod
    def evaluate(var: Variable, value: list[int] | int, precision: float = 1e-2) -> float:
        term = OneHot.encode(var, precision)
        binary_var = sorted(
            term.variables(),
            key=lambda x: _extract_number(x.label),
        )

        binary_list = OneHot._one_hot_encode(value, len(binary_var) + 1) if isinstance(value, int) else value

        if not OneHot.check_valid(binary_list)[0]:
            raise ValueError(f"invalid binary string {binary_list} with the one hot encoding.")

        # after encoding we will have one less variable than the binary list, because the first variable is multiplied
        # by 0 so it is removed from the term.
        if len(binary_list) < len(binary_var) + 1:
            for _ in range(len(binary_var) - len(binary_list) + 1):
                binary_list.append(0)
        elif len(binary_list) != len(binary_var) + 1:
            raise ValueError(f"expected {len(binary_var) + 1} variables but received {len(binary_list)}")

        zero_index = OneHot._find_zero(var)
        binary_dict: dict[BaseVariable, list[int]] = {}
        for i in range(var.num_binary_equivalent()):
            if i < zero_index:
                binary_dict[binary_var[i]] = [binary_list[i]]
            if i > zero_index:
                binary_dict[binary_var[i - 1]] = [binary_list[i]]

        _out = term.evaluate(binary_dict)

        if isinstance(_out, RealNumber):
            out = float(_out)
        elif isinstance(_out, complex) and _out.imag == 0:
            out = float(_out.real)
        else:
            raise ValueError(f"Evaluation answer ({_out}) is outside the variable domain ({var.domain}).")

        out = int(out) if var.domain in {Domain.INTEGER, Domain.POSITIVE_INTEGER} else out

        if not var.domain.check_value(out):
            raise ValueError(
                f"The value {out} violates the domain {var.domain.__class__.__name__} of the variable {var}"
            )  # not sure this line can be reached.

        return out

    @staticmethod
    def encoding_constraint(var: Variable, precision: float = 1e-2) -> ComparisonTerm:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0])) + 1

        binary_vars = [BinaryVariable(var.label + f"({i})") for i in range(n_binary)]
        return ComparisonTerm(lhs=sum(binary_vars), rhs=1, operation=ComparisonOperation.EQ)

    @staticmethod
    def num_binary_equivalent(var: Variable, precision: float = 1e-2) -> int:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0])) + 1

        return n_binary

    @staticmethod
    def check_valid(value: list[int] | int) -> tuple[bool, int]:
        binary_list = OneHot._one_hot_encode(value, value) if isinstance(value, int) else value
        num_ones = binary_list.count(1)
        return num_ones == 1, (num_ones - 1) ** 2


@yaml.register_class
class DomainWall(Encoding):
    """Represents a Domain-wall variable encoding class."""

    name = "Domain Wall"

    @staticmethod
    def _domain_wall_encode(x: int, N: int) -> list[int]:
        """domain wall encode integer x in range [0, N-1].

        Args:
            x (int): the value to be encoded
            N (int): the number of bits to encode x.

        Raises:
            ValueError: if x is larger than N or less than 0

        Returns:
            list[int]: a binary list representing the domain wall encoding of the integer x.
        """
        if not (0 <= x <= N):
            raise ValueError(f"the input value ({x}) must be in range [0, {N}]")
        return [1] * x + [0] * (N - x)

    @staticmethod
    def encode(var: Variable, precision: float = 1e-2) -> Term:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0]))

        binary_vars = [BinaryVariable(var.label + f"({i})") for i in range(n_binary)]

        term = Term([0], Operation.ADD)
        for i in range(n_binary):
            term += binary_vars[i]

        term += bounds[0]

        return term * var.precision if var.domain is Domain.REAL else term

    @staticmethod
    def evaluate(var: Variable, value: list[int] | int, precision: float = 1e-2) -> float:
        term = DomainWall.encode(var, precision)
        binary_var = term.variables()
        binary_var = sorted(
            term.variables(),
            key=lambda x: _extract_number(x.label),
        )

        binary_list: list[int] = (
            DomainWall._domain_wall_encode(value, len(binary_var)) if isinstance(value, RealNumber) else value
        )

        if not DomainWall.check_valid(binary_list)[0]:
            raise ValueError(f"invalid binary string {binary_list} with the domain wall encoding.")

        if len(binary_list) < len(binary_var):
            for _ in range(len(binary_var) - len(binary_list)):
                binary_list.append(0)
        elif len(binary_list) != len(binary_var):
            raise ValueError(f"expected {len(binary_var)} variables but received {len(binary_list)}")

        binary_dict: dict[BaseVariable, list[int]] = {binary_var[i]: [binary_list[i]] for i in range(len(binary_list))}

        _out = term.evaluate(binary_dict)

        if isinstance(_out, RealNumber):
            out = float(_out)
        elif isinstance(_out, complex) and _out.imag == 0:
            out = float(_out.real)
        else:
            raise ValueError(f"Evaluation answer ({_out}) is outside the variable domain ({var.domain}).")

        out = int(out) if var.domain in {Domain.INTEGER, Domain.POSITIVE_INTEGER} else out

        if not var.domain.check_value(out):
            raise ValueError(
                f"The value {out} violates the domain {var.domain.__class__.__name__} of the variable {var}"
            )  # not sure if this line is reachable.
        return out

    @staticmethod
    def encoding_constraint(var: Variable, precision: float = 1e-2) -> ComparisonTerm:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0]))

        binary_vars = [BinaryVariable(var.label + f"({i})") for i in range(n_binary)]
        return ComparisonTerm(
            lhs=sum(binary_vars[i + 1] * (1 - binary_vars[i]) for i in range(len(binary_vars) - 1)),
            rhs=0,
            operation=ComparisonOperation.EQ,
        )

    @staticmethod
    def num_binary_equivalent(var: Variable, precision: float = 1e-2) -> int:
        bounds = var.bounds
        if var.domain is Domain.REAL:
            bounds = (bounds[0] / precision, bounds[1] / precision)

        n_binary = int(np.abs(bounds[1] - bounds[0]))

        return n_binary

    @staticmethod
    def check_valid(value: list[int] | int) -> tuple[bool, int]:
        binary_list = DomainWall._domain_wall_encode(value, value) if isinstance(value, int) else value
        value = sum(binary_list[i + 1] * (1 - binary_list[i]) for i in range(len(binary_list) - 1))
        return value == 0, value


# Variables ###


class BaseVariable(ABC):
    """
    Abstract base class for symbolic decision variables.
    """

    def __init__(self, label: str, domain: Domain, bounds: tuple[float | None, float | None] = (None, None)) -> None:
        """initialize a new Variable object

        Args:
            label (str): The name of the variable.
            domain (Domain): The domain of the values this variable can take.
            bounds (tuple[float  |  None, float  |  None], optional): the bounds on the variable's values.
                                                The bounds follow the structure (lower_bound, Upper_bound) both
                                                included. Defaults to (None, None).
                                                Note: if None is selected then the lowest/highest possible value of the
                                                variable's domain is chosen.

        Raises:
            OutOfBoundsException: the lower bound or the upper bound don't correspond to the variable domain.
            InvalidBoundsError: the lower bound is higher than the upper bound.
        """
        self._label = label
        self._domain = domain

        lower_bound, upper_bound = bounds
        if lower_bound is None:
            lower_bound = domain.min()
        if upper_bound is None:
            upper_bound = domain.max()

        if not self.domain.check_value(upper_bound):
            raise OutOfBoundsException(
                f"the lower bound ({upper_bound}) does not respect the domain of the variable ({self.domain})"
            )
        if not self.domain.check_value(lower_bound):
            raise OutOfBoundsException(
                f"the upper bound ({lower_bound}) does not respect the domain of the variable ({self.domain})"
            )
        if lower_bound > upper_bound:
            raise InvalidBoundsError("lower bound can't be larger than the upper bound.")
        self._bounds = (lower_bound, upper_bound)

    @property
    def bounds(self) -> tuple[float, float]:
        """Property that stores a tuple representing the bounds of the values a variable is allowed to take.º

        Returns:
            tuple(float, float): The lower and upper bound of the variable.
        """
        return self._bounds

    @property
    def lower_bound(self) -> float:
        """The lower bound of the variable.

        Returns:
            float: the value of the lower bound.
        """
        return self._bounds[0]

    @property
    def upper_bound(self) -> float:
        """The upper bound of the variable.

        Returns:
            float: the value of the upper bound.
        """
        return self._bounds[1]

    @property
    def label(self) -> str:
        """the label (name) of the variable.

        Returns:
            string: the name of the variable.
        """
        return self._label

    @property
    def domain(self) -> Domain:
        """The domain of values that the variable is allowed to take.

        Returns:
            Domain: The domain of the values the variable can take.
        """
        return self._domain

    def set_bounds(self, lower_bound: float | None, upper_bound: float | None) -> None:
        """set the bounds of the variable.

        Args:
            lower_bound (float | None): The lower bound (if None the lowest allowed bound in the variable domain is
            selected). Defaults to None.
            upper_bound (float | None): The upper bound (if None the highest allowed bound in the variable domain is
            selected). Defaults to None.
        Raises:
            OutOfBoundsException: the lower bound or the upper bound don't correspond to the variable domain.
            InvalidBoundsError: the lower bound is higher than the upper bound.
        """
        if lower_bound is None:
            lower_bound = self._domain.min()
        if upper_bound is None:
            upper_bound = self._domain.max()
        if not self.domain.check_value(lower_bound):
            raise OutOfBoundsException(
                f"the lower bound ({lower_bound}) does not respect the domain of the variable ({self.domain})"
            )
        if not self.domain.check_value(upper_bound):
            raise OutOfBoundsException(
                f"the upper bound ({upper_bound}) does not respect the domain of the variable ({self.domain})"
            )
        if lower_bound > upper_bound:
            raise InvalidBoundsError(
                f"the lower bound ({lower_bound}) should not be greater than the upper bound ({upper_bound})"
            )
        self._bounds = (lower_bound, upper_bound)

    @abstractmethod
    def num_binary_equivalent(self) -> int:
        """
        Returns:
            int: the number of binary variables that are needed to represent this variable in the given encoding.
        """

    @abstractmethod
    def evaluate(self, value: list[int] | RealNumber) -> RealNumber:
        """Evaluates the value of the variable given a binary string or a number.

        Args:
            value (list[int] | int | float): the value used to evaluate the variable.
                If the value provided is binary list (list[int]) then the value of the variable is evaluated based on
                its binary representation. This representation is constructed using the encoding, bounds and domain
                of the variable. To check the binary representation of a variable you can check the method `to_binary()`

        Returns:
            int | float | complex: the evaluated vale of the variable.
        """

    def update_variable(self, domain: Domain, bounds: tuple[float | None, float | None] = (None, None)) -> None:
        """Replaces the information of the variable with those coming from the dictionary
        if the variable label is in the dictionary

        Args:
            domain (Domain): The updated domain of the variable.
            bounds (tuple[float | None, float | None]): The updated bounds of the variable. Defaults to (None, None)
        """

        self._domain = domain
        self.set_bounds(bounds[0], bounds[1])

    @abstractmethod
    def to_binary(self) -> Term:
        """Returns the binary representation of a variable.

        Returns:
            Term: the binary representation of a variable.
        """

    def __repr__(self) -> str:
        return f"{self._label}"

    def __str__(self) -> str:
        return f"{self._label}"

    def __add__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        if isinstance(other, Term):
            return other + self

        if isinstance(other, np.generic):
            other = other.item()

        return Term(elements=[self, other], operation=Operation.ADD)

    __radd__ = __add__
    __iadd__ = __add__

    def __mul__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        if isinstance(other, Term):
            return other * self

        if isinstance(other, np.generic):
            other = other.item()

        return Term(elements=[self, other], operation=Operation.MUL)

    def __rmul__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        if isinstance(other, Term):
            return other * self

        if isinstance(other, np.generic):
            other = other.item()

        return Term(elements=[other, self], operation=Operation.MUL)

    __imul__ = __mul__

    def __sub__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented

        if isinstance(other, np.generic):
            other = other.item()

        return self + -1 * other

    def __rsub__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented

        if isinstance(other, np.generic):
            other = other.item()

        return -1 * self + other

    __isub__ = __sub__

    def __neg__(self) -> Term:
        return -1 * self

    def __truediv__(self, other: RealNumber) -> Term:
        if not isinstance(other, RealNumber):
            raise NotImplementedError("Only division by real numbers is currently supported")

        if other == 0:
            raise ValueError("Division by zero is not allowed")

        if isinstance(other, np.generic):
            other = other.item()
        other = 1 / other
        return self * other

    __itruediv__ = __truediv__

    def __rtruediv__(self, other: Number | BaseVariable | Term) -> Term:
        raise NotSupportedOperation("Only division by numbers is currently supported")

    def __rfloordiv__(self, other: Number | BaseVariable | Term) -> Term:
        raise NotSupportedOperation("Only division by numbers is currently supported")

    def __pow__(self, a: int) -> Term:
        out: BaseVariable | Term = copy.copy(self)

        if a < 0:
            raise NotImplementedError("Negative Power is not Supported.")

        if a == 0:
            return Term(elements=[1], operation=Operation.ADD)

        for _ in range(a - 1):
            out *= copy.copy(self)

        if isinstance(out, BaseVariable):
            out = Term(elements=[out], operation=Operation.ADD)
        return out

    def __hash__(self) -> int:
        return hash((self._label, self._domain.value, self._bounds))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseVariable):
            return False
        return hash(self) == hash(other)


@yaml.register_class
class BinaryVariable(BaseVariable):
    """
    Binary decision variable restricted to the set ``{0, 1}``.

    Example:
        .. code-block:: python

            from qilisdk.core.variables import BinaryVariable

            x = BinaryVariable("x")
    """

    def __init__(self, label: str) -> None:
        super().__init__(label=label, domain=Domain.BINARY)

    def num_binary_equivalent(self) -> int:  # noqa: PLR6301
        return 1

    def evaluate(self, value: list[int] | RealNumber) -> RealNumber:
        if isinstance(value, int | float):
            if value in {1.0, 0.0}:
                return int(value)
            if not self.domain.check_value(value):
                raise EvaluationError(f"Evaluating a Binary variable with a value {value} that is outside the domain.")
            return value  # I don't think this line is reachable
        if len(value) != 1:
            raise EvaluationError("Evaluating a Binary variable with a binary list of more than one item.")
        return value[0]

    def update_variable(self, domain: Domain, bounds: tuple[float | None, float | None] = (None, None)) -> None:
        raise NotImplementedError

    def to_binary(self) -> Term:
        return Term([self], Operation.ADD)

    def __copy__(self) -> BinaryVariable:
        return BinaryVariable(label=self.label)


@yaml.register_class
class SpinVariable(BaseVariable):
    """Represents Spin Variable structure."""

    def __init__(self, label: str) -> None:
        super().__init__(label=label, domain=Domain.SPIN, bounds=(-1, 1))

    def num_binary_equivalent(self) -> int:  # noqa: PLR6301
        return 1

    def update_variable(self, domain: Domain, bounds: tuple[float | None, float | None] = (None, None)) -> None:
        raise NotImplementedError

    def evaluate(self, value: list[int] | RealNumber) -> RealNumber:
        if isinstance(value, Number):
            if not self.domain.check_value(value) and value != 0:
                raise EvaluationError(f"Evaluating a Spin variable with a value {value} that is outside the domain.")
            return -1 if value in {0, -1} else 1
        if len(value) != 1:
            raise EvaluationError("Evaluating a Spin variable with a list of more than one item.")
        return -1 if value[0] in {0, -1} else 1

    def to_binary(self) -> Term:
        return Term([self], Operation.ADD)

    def __copy__(self) -> SpinVariable:
        return SpinVariable(label=self.label)


@yaml.register_class
class Variable(BaseVariable):
    """
    Generic (possibly continuous) optimization variable with configurable encoding.

    Example:
        .. code-block:: python

            from qilisdk.core.variables import Domain, Variable

            price = Variable("price", domain=Domain.REAL, bounds=(0, 10))
            binary_term = price.to_binary()
    """

    def __init__(
        self,
        label: str,
        domain: Domain,
        bounds: tuple[float | None, float | None] = (None, None),
        encoding: type[Encoding] = Bitwise,
        precision: float = 1e-2,
    ) -> None:
        """

        Args:
            label (str): The name of the variable.
            domain (Domain): The domain of the values this variable can take.
            bounds (tuple[float  |  None, float  |  None], optional): the bounds on the values of the variable The bounds
                    have the structure (lower_bound, Upper_bound) both values included. Defaults to (None, None).
                    Note: if None is selected then the lowest/highest possible value of the variable's domain is chosen.
            encoding (type[Encoding], optional): _description_. Defaults to Bitwise.
            precision (float, optional): The floating point precision for REAL variables. Defaults to 1e-2.
        """
        super().__init__(label=label, domain=domain, bounds=bounds)
        self._encoding = encoding
        self._precision = 1e-2
        self._term: Term | None = None
        self._bin_vars: list[BaseVariable] = []
        self._precision = precision

    @property
    def encoding(self) -> type[Encoding]:
        return self._encoding

    @property
    def precision(self) -> float:
        return self._precision

    @property
    def term(self) -> Term:
        if self._term is None:
            if self.bounds[1] > LARGE_BOUND or self.bounds[0] < -LARGE_BOUND:
                logger.warning(
                    f"Encoding variable {self.label} which has the bounds {self.bounds}"
                    + "is very expensive and may take a very long time."
                )
            self._term = self.to_binary()
        return self._term

    @property
    def bin_vars(self) -> list[BaseVariable]:
        if self._term is None:
            self.to_binary()
        return self._bin_vars

    def set_precision(self, precision: float) -> None:
        self._precision = precision
        self._term = None

    def __copy__(self) -> Variable:
        return Variable(label=self.label, domain=self.domain, bounds=self.bounds, encoding=self._encoding)

    def __getitem__(self, item: int) -> BaseVariable:
        if self._term is None:
            self.to_binary()
        return self._bin_vars[item]

    def update_variable(
        self,
        domain: Domain,
        bounds: tuple[float | None, float | None] = (None, None),
        encoding: type[Encoding] | None = None,
    ) -> None:
        self._encoding = encoding if encoding is not None else self._encoding
        self._term = None
        return super().update_variable(domain, bounds)

    def evaluate(self, value: list[int] | RealNumber) -> RealNumber:
        if isinstance(value, int | float):
            if not self.domain.check_value(value):
                raise ValueError(f"The value {value} is invalid for the domain {self.domain.value}")
            if value < self.lower_bound or value > self.upper_bound:
                raise ValueError(f"The value {value} is outside the defined bounds {self.bounds}")
            return value
        return self.encoding.evaluate(self, value, self._precision)

    def to_binary(self) -> Term:
        if self._term is None:
            term = self.encoding.encode(self, precision=self._precision)
            self._term = copy.copy(term)
            self._bin_vars = [BinaryVariable(f"{self.label}({i})") for i in range(self.num_binary_equivalent())]
            self._bin_vars = sorted(
                self._bin_vars,
                key=lambda x: _extract_number(x.label),
            )
        return self._term

    def num_binary_equivalent(self) -> int:
        """
        Returns:
            int: the number of binary variables needed to encode the continuous variable.
        """
        return self.encoding.num_binary_equivalent(self, precision=self._precision)

    def check_valid(self, binary_list: list[int]) -> tuple[bool, int]:
        """checks if the binary list sample is a valid sample in the variable's encoding.

        Args:
            binary_list (list[int] | int):  a list of binary values or an integer value.

        Returns:
            tuple[bool, int]: the boolean is True if the sample is a valid encoding,
                                and the integer is the error in the encoding.
        """
        return self.encoding.check_valid(binary_list)

    def encoding_constraint(self) -> ComparisonTerm:
        """Given a continuous variable return a Comparison Term that ensures that the encoding is respected.

        Returns:
            ComparisonTerm: a Comparison Term that ensures the encoding is respected.
        """
        return self.encoding.encoding_constraint(self, precision=self._precision)


@yaml.register_class
class Parameter(BaseVariable):
    """
    Symbolic scalar used to parametrize expressions while remaining differentiable.

    Example:
        .. code-block:: python

            from qilisdk.core.variables import Parameter

            theta = Parameter("theta", value=0.5)
            theta.set_value(0.75)
    """

    def __init__(
        self,
        label: str,
        value: RealNumber,
        domain: Domain = Domain.REAL,
        bounds: tuple[float | None, float | None] = (None, None),
    ) -> None:
        super().__init__(label=label, domain=domain, bounds=bounds)

        if not self.domain.check_value(value):
            raise ValueError(
                f"Parameter value provided ({value}) doesn't correspond to the parameter's domain ({self.domain.name})"
            )
        self._value = value
        self.set_bounds(bounds[0], bounds[1])

    @property
    def value(self) -> RealNumber:
        return self._value

    def set_value(self, value: RealNumber) -> None:
        if not self.domain.check_value(value):
            raise ValueError(
                f"Parameter value provided ({value}) doesn't correspond to the parameter's domain ({self.domain.name})"
            )
        if value > self.bounds[1] or value < self.bounds[0]:
            raise ValueError(f"The value provided ({value}) is outside the bound of the parameter {self.bounds}")
        self._value = value

    def num_binary_equivalent(self) -> int:  # noqa: PLR6301
        """
        Returns:
        int: the number of binary variables that are needed to represent this variable in the given encoding.
        """
        return 0

    def evaluate(self, value: list[int] | RealNumber = 0) -> RealNumber:
        """Evaluates the value of the variable given a binary string or a number.

        Args:
            value (list[int] | int | float): the value used to evaluate the variable.
                If the value provided is binary list (list[int]) then the value of the variable is evaluated based on
                its binary representation. This representation is constructed using the encoding, bounds and domain
                of the variable. To check the binary representation of a variable you can check the method `to_binary()`

        Returns:
            float: the evaluated vale of the variable.
        """
        return self.value

    def to_binary(self) -> Term:
        """Returns the binary representation of a variable.

        Returns:
            Term: the binary representation of a variable.
        """
        return Term([self.value], operation=Operation.ADD)

    def set_bounds(self, lower_bound: float | None, upper_bound: float | None) -> None:
        upper_bound = upper_bound if upper_bound is not None else self.domain.max()
        lower_bound = lower_bound if lower_bound is not None else self.domain.min()
        if self.value > upper_bound or self.value < lower_bound:
            raise ValueError(
                f"The current value of the parameter ({self.value}) is outside the bounds ({lower_bound}, {upper_bound})"
            )
        super().set_bounds(lower_bound, upper_bound)

    def update_variable(self, domain: Domain, bounds: tuple[float | None, float | None] = (None, None)) -> None:
        if len(bounds) != 2:  # noqa: PLR2004
            raise ValueError(
                "Invalid bounds provided: the bounds need to be a tuple with the format (lowe_bound, upper_bound)"
            )

        if domain.check_value(self.value):
            self._domain = domain
        else:
            raise ValueError(
                f"The provided domain ({domain.name}) is incompatible with the current parameter value ({self.value})"
            )

        self.set_bounds(lower_bound=bounds[0], upper_bound=bounds[1])


# Terms ###


@yaml.register_class
class Term:
    """Represents a mathematical Term (e.g. 3x*y, 2x, ...).

    And they are built from:
    - ``Variable``'s: The decision variables of the model (x, y, ...).
    - Other ``Term``'s: Allowing for complex expressions to be constructed.
    """

    CONST = Variable(CONST_KEY, Domain.REAL)

    def __init__(self, elements: Sequence[BaseVariable | Term | Number], operation: Operation) -> None:
        """initialize a new term object.

        Args:
            elements (Sequence[BaseVariable  |  Term  |  Number]): a list of elements in the term.
            operation (Operation): the mathematical operation between these elements.

        Raises:
            ValueError: if the items inside elements are not from the listed types (BaseVariable  |  Term  |  Number).
        """
        self._operation = operation
        self._elements: dict[BaseVariable | Term, Number] = {}  # The list of elements in the term.
        # key: the term or variable | value: the coefficient corresponding to that value.
        for e in elements:
            if isinstance(e, BaseVariable):
                if e in self:
                    if self._is_constant(e):
                        self[e] = self._apply_operation_on_constants([self[e], 1])
                    elif isinstance(e, BinaryVariable) and self.operation == Operation.MUL:
                        self[e] = 1
                    else:
                        self[e] += 1
                else:
                    self[e] = 1
            elif isinstance(e, Number):
                if self.CONST in self:
                    self[self.CONST] = self._apply_operation_on_constants([self[self.CONST], e])
                else:
                    self[self.CONST] = e
            elif isinstance(e, Term):
                if len(e) == 0:
                    if self.CONST in self:
                        self[self.CONST] = self._apply_operation_on_constants([self[self.CONST], 0])
                    else:
                        self[self.CONST] = 0
                elif e.operation == self._operation:
                    for key in e:
                        if key in self:
                            if isinstance(key, BaseVariable) and self._is_constant(key):
                                self[key] = self._apply_operation_on_constants([self[key], e[key]])
                            elif isinstance(key, BinaryVariable) and self.operation == Operation.MUL:
                                self[key] = 1
                            else:
                                self[key] += e[key]
                        else:
                            self[key] = e[key]
                else:
                    e_copy = copy.copy(e)
                    coeff = complex(1.0)
                    if e_copy.operation == Operation.MUL and self.CONST in e_copy:
                        coeff = e_copy.pop(self.CONST)
                    simple_e = e_copy._simplify()  # noqa: SLF001
                    simple_e = self.CONST if isinstance(simple_e, Term) and len(simple_e) == 0 else simple_e
                    if simple_e in self:
                        if isinstance(simple_e, BaseVariable) and self._is_constant(simple_e):
                            self[simple_e] = self._apply_operation_on_constants([self[simple_e], coeff])
                        else:
                            self[simple_e] += coeff
                    else:
                        self[simple_e] = coeff
            else:
                raise ValueError(
                    f"Term accepts object of types Term or Variable but an object of type {e.__class__} was given"
                )
        self._remove_zeros()

    @property
    def operation(self) -> Operation:
        """
        Returns:
            Operation: the operation between the term's elements.
        """
        return self._operation

    @property
    def degree(self) -> int:
        """
        Returns:
            int: the highest degree in the term.
        """
        degree = 0
        if self.operation == Operation.MUL:
            for element in self:
                if isinstance(element, Term):
                    degree += element.degree
                elif isinstance(element, BaseVariable) and not self._is_constant(element):
                    degree += int(_assert_real(self[element]))
            return degree

        for element in self:
            if isinstance(element, Term):
                degree = max(degree, element.degree)
            elif isinstance(element, BaseVariable) and not self._is_constant(element):
                degree = max(degree, 1)
        return degree

    def to_binary(self) -> Term:
        """Returns the term in binary format. That is encoding all continuous variables into
            binary according to the encoding defined in the variable.

        Raises:
            ValueError: The term contains operations that are not addition or multiplication.
            ValueError: the term contains an element that is not a Term or a BaseVariable.

        Returns:
            Term: the term after transforming all the variables into binary.
        """
        if self.operation not in {Operation.ADD, Operation.MUL}:
            raise ValueError("Can not evaluate any operation that is not Addition of Multiplication")
        out_list: list[BaseVariable | Term | Number] = []
        for e in self:
            if isinstance(e, Term):
                out_list.append(self[e] * e.to_binary())
            elif isinstance(e, BaseVariable):
                if self._is_constant(e):
                    out_list.append(self[e])
                elif isinstance(e, Variable):
                    x = e.to_binary()
                    if self.operation == Operation.MUL:
                        out_list.append(x ** int(_assert_real(self[e])))
                    else:
                        out_list.append(self[e] * x)
                else:
                    out_list.append(self[e] * e)
            else:
                raise ValueError(f"Evaluating term with elements of type {e.__class__} is not supported.")

        return Term(out_list, self.operation)

    def _apply_operation_on_constants(self, const_list: list[Number]) -> Number:
        out = complex(const_list[0])
        for i in range(1, len(const_list)):
            if self.operation is Operation.ADD:
                out += const_list[i]
            elif self.operation is Operation.SUB:
                out -= const_list[i]
            elif self.operation is Operation.MUL:
                out *= const_list[i]
            elif self.operation is Operation.DIV:
                out /= const_list[i]

        return out

    def variables(self) -> list[BaseVariable]:
        """Returns the unique list of variables in the Term

        Returns:
            list[Variable]: The unique list of variables in the Term.
        """
        var = set()
        for e in self:
            if isinstance(e, BaseVariable) and not self._is_constant(e):
                var.add(e)
            elif isinstance(e, Term):
                var.update(e.variables())
        return sorted(var, key=lambda x: x.label)

    def _simplify(self) -> Term | BaseVariable:
        """Simplify the term object.

        Returns:
            (Term | BaseVariable): the simplified term.
        """
        if len(self) == 1:
            item = next(iter(self._elements.keys()))
            if self._elements[item] == 1:
                return item
        return self

    def pop(self, item: BaseVariable | Term) -> Number:
        """Remove an item from the term.

        Args:
            item (BaseVariable | Term): the item to be removed.

        Raises:
            KeyError: if item is not in the term.

        Returns:
            Number: the coefficient of the removed item.
        """
        try:
            return self._elements.pop(item)
        except KeyError as e:
            raise KeyError(f'item "{item}" not found in the term.') from e

    def _is_constant(self, variable: BaseVariable) -> bool:
        """Checks if the variable is a constant variable as defined by the Term class.

        Args:
            variable (BaseVariable): the variable to be checked.

        Returns:
            bool: True if the variable is a constant, False otherwise.
        """
        return variable == self.CONST

    def to_list(self) -> list[BaseVariable | Term | Number]:
        """Exports the current term into a list of its elements.

        Returns:
            list[BaseVariable | Term | Number]: A list of the elements inside the term.
        """
        out_list: list[BaseVariable | Term | Number] = []
        for e in self:
            if isinstance(e, BaseVariable) and self._is_constant(e):
                out_list.append(self[e])
            elif self.operation == Operation.MUL:
                for _ in range(int(_assert_real(self[e]))):
                    out_list.append(e)
            else:
                out_list.append(self[e] * e if self[e] != 1 else e)
        return out_list

    def _unfold_parentheses(self) -> Term:
        """Simplifies any parentheses in the term expression.

        Returns:
            Term: A new term with a more simplified form.
        """
        out = copy.copy(self)
        if out.operation != Operation.MUL:
            return out

        parentheses: list[tuple[Term, Number]] = []

        for e in out:
            if isinstance(e, Term) and e.operation == Operation.ADD:
                parentheses.append((copy.copy(e), out[e]))

        for term, _ in parentheses:
            out.pop(term)

        if len(out) == 0 and len(parentheses) != 0:
            out = Term([1], Operation.ADD)

        for _term, coeff in parentheses:
            term = copy.copy(_term)
            _coeff = _assert_real(coeff)
            if _coeff > 1:
                term **= int(_coeff)
            final_out = []
            for t in term:
                final_out.append(t * out * term[t])
            out = Term(final_out, Operation.ADD)

        return out

    def _remove_zeros(self) -> None:
        """Simplifies any un-necessary zeros from terms."""
        to_be_popped = []
        if self.operation == Operation.MUL and self.CONST in self and self[self.CONST] == 0:
            l = len(self)
            for _ in range(l):
                self._elements.popitem()
        for e in self:
            if self[e] == 0:
                to_be_popped.append(e)
        for p in to_be_popped:
            self._elements.pop(p)

    def evaluate(self, var_values: Mapping[BaseVariable, list[int] | RealNumber]) -> Number:
        """Evaluates the term given a set of values for the variables in the term.

        Args:
            var_values (Mapping[BaseVariable, list[int]  |  Number]): the values of the variables in the term.
                If the value provided is binary list (list[int]) then the value of the variable is evaluated based on
                its binary representation. This representation is constructed using the encoding, bounds and domain
                of the variable. To check the binary representation of a variable you can check the method `to_binary()`

        Raises:
            ValueError: if not all variables in the term are provided a value.

        Returns:
            float: the result from evaluating the term.
        """
        if len(self._elements) == 0:
            return 0
        _var_values = dict(var_values)
        for var in self.variables():
            if isinstance(var, Parameter):
                _var_values[var] = var.value
            if var not in _var_values:
                raise ValueError(f"Can not evaluate term because the value of the variable {var} is not provided.")
        output = complex(0.0) if self.operation in {Operation.ADD, Operation.SUB} else complex(1.0)
        for e in self:
            if isinstance(e, Term):
                output = self._apply_operation_on_constants([output, e.evaluate(_var_values) * self[e]])
            elif isinstance(e, BaseVariable):
                if e == self.CONST:
                    output = self._apply_operation_on_constants([output, self[e]])
                elif self.operation == Operation.MUL:
                    output = self._apply_operation_on_constants([output, e.evaluate(_var_values[e]) ** self[e]])
                else:
                    output = self._apply_operation_on_constants([output, e.evaluate(_var_values[e]) * self[e]])
        if isinstance(output, RealNumber):
            return float(output)
        if isinstance(output, complex) and output.imag == 0:
            return float(output.real)
        return output

    def get_constant(self) -> Number:
        """
        Returns:
            Number: The constant value of the term.
        """
        if self.CONST in self:
            return self[self.CONST]
        return 0 if self.operation in {Operation.ADD, Operation.SUB} else 1

    def is_parameterized_term(self) -> bool:
        return all(isinstance(var, Parameter) for var in self.variables())

    def __copy__(self) -> Term:
        return Term(copy.copy(self.to_list()), copy.copy(self.operation))

    def __repr__(self) -> str:
        if len(self) == 0:
            return "0"
        output_string = ""
        const = self.get_constant()
        keys = list(self._elements.keys())

        if (
            (self.operation in {Operation.ADD, Operation.SUB} and const == 0)
            or (self.operation in {Operation.MUL, Operation.DIV} and const == 1)
        ) and Term.CONST in keys:
            keys.remove(Term.CONST)

        for i, e in enumerate(keys):
            if isinstance(e, Term):
                term_str = str(e).strip()
                if len(term_str) > 0:
                    if term_str[0] == "(" and term_str[-1] == ")":
                        term_str = term_str.removeprefix("(").removesuffix(")")
                    output_string += (
                        f"({term_str}) " if self[e] == 1 else f"({_float_if_real(self[e])}) * ({term_str}) "
                    )
            elif isinstance(e, BaseVariable):
                if self._is_constant(e):
                    if self.operation in {Operation.ADD, Operation.SUB} and self[e] == 0:
                        continue
                    if self.operation in {Operation.MUL, Operation.DIV} and self[e] == 1:
                        continue
                    output_string += f"({_float_if_real(self[e])}) "
                elif (self.operation is Operation.MUL or self.operation is Operation.DIV) and _assert_real(self[e]) > 1:
                    output_string += f"({e}^{_float_if_real(self[e])}) "
                else:
                    output_string += f"{e} " if self[e] == 1 else f"({_float_if_real(self[e])}) * {e} "
            else:
                output_string += f"{e} "
            if i < len(keys) - 1:
                output_string += f"{self.operation.value} "

        return output_string.strip()

    __str__ = __repr__

    def __getitem__(self, item: BaseVariable | Term) -> Number:
        return self._elements[item]

    def __setitem__(self, key: BaseVariable | Term, item: Number) -> None:
        self._elements[key] = item

    def __iter__(self) -> Iterator[BaseVariable | Term]:
        yield from self._elements

    def __contains__(self, item: BaseVariable | Term) -> bool:
        return item in self._elements

    __next__ = __iter__

    def __len__(self) -> int:
        return len(self._elements)

    def __add__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        out = self.to_list() if self.operation == Operation.ADD else [copy.copy(self)]

        if isinstance(other, np.generic):
            other = other.item()

        out.append(other)
        return Term(out, Operation.ADD)

    __iadd__ = __add__

    def __radd__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        out = self.to_list() if self.operation == Operation.ADD else [copy.copy(self)]

        if isinstance(other, np.generic):
            other = other.item()
        out.insert(0, other)
        return Term(out, Operation.ADD)

    def __mul__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        out = self.to_list() if self.operation == Operation.MUL else [copy.copy(self)]
        if len(out) == 0:
            out = [0]

        if isinstance(other, np.generic):
            other = other.item()

        out.append(other)
        return Term(out, Operation.MUL)._unfold_parentheses()

    __imul__ = __mul__

    def __rmul__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        out = self.to_list() if self.operation == Operation.MUL else [copy.copy(self)]
        if len(out) == 0:
            out = [0]

        if isinstance(other, np.generic):
            other = other.item()

        out.insert(0, other)
        return Term(out, Operation.MUL)._unfold_parentheses()

    def __neg__(self) -> Term:
        return -1 * self

    def __sub__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented

        if isinstance(other, np.generic):
            other = other.item()

        return self + -1 * other

    def __rsub__(self, other: Number | BaseVariable | Term) -> Term:
        if not isinstance(other, (Number, BaseVariable, Term)):
            return NotImplemented
        return -1 * self + other

    __isub__ = __sub__

    def __truediv__(self, other: Number) -> Term:
        if not isinstance(other, Number):
            raise NotImplementedError("Only division by numbers is currently supported")

        if other == 0:
            raise ValueError("Division by zero is not allowed")

        other = 1 / other
        return self * other

    __itruediv__ = __truediv__

    def __rtruediv__(self, other: Number | BaseVariable | Term) -> Term:
        raise NotSupportedOperation("Only division by numbers is currently supported")

    def __rfloordiv__(self, other: Number | BaseVariable | Term) -> Term:
        raise NotSupportedOperation("Only division by numbers is currently supported")

    def __pow__(self, a: int) -> Term:
        if not isinstance(a, int):
            raise ValueError(f"Only integer exponents are allowed, but provided {type(a)}")
        if self.operation == Operation.ADD:
            out = copy.copy(self)
            for _ in range(a - 1):
                out_list = []
                for element in self:
                    out_list.append(out * copy.copy(element) * self[element])
                out = Term(out_list, Operation.ADD)
            return out
        if self.operation == Operation.MUL:
            out = copy.copy(self)
            for element in out:
                if element is Term.CONST:
                    out[element] **= a
                else:
                    out[element] *= a
            return out
        raise NotImplementedError(
            "The power operation for terms that are not addition or multiplication is not supported."
        )

    def __hash__(self) -> int:
        return hash((frozenset(self._elements.items()), self.operation))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Term):
            return False
        return hash(self) == hash(other)


@yaml.register_class
class ComparisonTerm:
    """Represents a mathematical comparison Term, that can be an equality or an inequality between two ``Term`` objects
    (e.g. x+y>0, x>2, ...).

    They are built from a left and a right hand part, each of which can contain:
    - ``Variable``'s: The decision variables of the model (x, y, ...).
    - Other ``Term``'s: Allowing for complex expressions to be constructed (x+y, ...)
    """

    def __init__(
        self,
        lhs: RealNumber | BaseVariable | Term,
        rhs: RealNumber | BaseVariable | Term,
        operation: ComparisonOperation,
    ) -> None:
        """Initializes a new comparison term.

        Args:
            lhs (RealNumber | BaseVariable | Term): the left hand side of the comparison term.
            rhs (RealNumber | BaseVariable | Term): the right hand side of the comparison term.
            operation (ComparisonOperation): the comparison operations between the left and right hand sides.
        """
        term = lhs - rhs
        if not isinstance(term, Term):
            term = Term([term], Operation.ADD)  # I don't think this line is reachable
        const = -1 * term.pop(Term.CONST) if Term.CONST in term else 0
        self._lhs = term
        self._rhs = Term([const], Operation.ADD)
        self._operation = operation

    @property
    def operation(self) -> ComparisonOperation:
        """
        Returns:
            ComparisonOperation: the comparison operation between the left and right hand sides.
        """
        return self._operation

    @property
    def lhs(self) -> Term:
        """
        Returns:
            Term: the left hand side of the comparison term.
        """
        return self._lhs

    @property
    def rhs(self) -> Term:
        """
        Returns:
            Term: the right hand side of the comparison term.
        """
        return self._rhs

    def variables(self) -> list[BaseVariable]:
        """Returns the unique list of variables in the Term

        Returns:
            list[Variable]: The unique list of variables in the Term.
        """
        lhs_var = self._lhs.variables()
        rhs_var = self._rhs.variables()

        var = set()
        var.update(lhs_var)
        var.update(rhs_var)

        return sorted(var, key=lambda x: x.label)

    @property
    def degree(self) -> int:
        """
        Returns:
            int: the maximum degree in the left and right hand sides of the comparison term.
        """
        return max(self.rhs.degree, self.lhs.degree)

    def to_list(self) -> list:
        """Exports the comparison term into a list. The elements of the right hand side are first moved to the left hand
        side before the generation of the list. Therefore, you can assume that the right hand side will be zero.

        Returns:
            list: a list constructed from all the elements in the left and right hand sides of the comparison term.
        """
        logger.info(
            "to_list(): The elements of output list assume the comparison term has been transformed "
            + f"from (lhs {self.operation.value} rhs) to (lhs - rhs {self.operation.value} 0).",
        )
        out = self.lhs.to_list()
        out.extend((-1 * self.rhs).to_list())
        return out

    def to_binary(self) -> ComparisonTerm:
        """Returns the comparison term in binary format. That is encoding all continuous variables into
            binary according to the encoding defined in the variable.

        Returns:
            ComparisonTerm: the comparison term after transforming all the variables into binary.
        """
        return ComparisonTerm(rhs=self.rhs.to_binary(), lhs=self.lhs.to_binary(), operation=self.operation)

    def _apply_comparison_operation(self, v1: RealNumber, v2: RealNumber) -> bool:
        """Compare two arguments.

        Args:
            v1 (Number): the left hand side value.
            v2 (Number): the right hand side value.

        Raises:
            ValueError: if the comparison term's operation is invalid.

        Returns:
            bool: the result of the comparison between v1 and v2 assuming the
            comparison operation of the comparison term object.
        """
        if self.operation is ComparisonOperation.EQ:
            return v1 == v2
        if self.operation is ComparisonOperation.GEQ:
            return v1 >= v2
        if self.operation is ComparisonOperation.GT:
            return v1 > v2
        if self.operation is ComparisonOperation.LEQ:
            return v1 <= v2
        if self.operation is ComparisonOperation.LT:
            return v1 < v2
        if self.operation is ComparisonOperation.NEQ:
            return v1 != v2
        raise ValueError(f"Unsupported Operation of type {self.operation.value}")

    def evaluate(self, var_values: Mapping[BaseVariable, RealNumber | list[int]]) -> bool:
        """Evaluates the comparison term given a set of values for the variables in the term.

        Args:
            var_values (Mapping[BaseVariable, list[int]  |  RealNumber]): the values of the variables in the comparison term.

        Returns:
            bool: the result from evaluating the comparison term.

        Raises:
            ValueError: if the constraint contains imaginary numbers.
        """
        lhs = self._lhs.evaluate(var_values)
        rhs = self._rhs.evaluate(var_values)
        if isinstance(lhs, complex):
            if lhs.imag != 0:
                raise ValueError("evaluating inequality constraints with complex values is not allowed")
            lhs = lhs.real
        if isinstance(rhs, complex):
            if rhs.imag != 0:
                raise ValueError("evaluating inequality constraints with complex values is not allowed")
            rhs = rhs.real
        return self._apply_comparison_operation(lhs, rhs)

    def __copy__(self) -> ComparisonTerm:
        return ComparisonTerm(rhs=copy.copy(self.rhs), lhs=copy.copy(self.lhs), operation=self.operation)

    def __repr__(self) -> str:
        return f"{str(self.lhs).strip()} {self.operation.value} {str(self.rhs).strip()}"

    __str__ = __repr__

    def __bool__(self) -> bool:
        raise TypeError(
            "Symbolic Constraint Term objects do not have an inherent truth value. "
            "Use a method like .evaluate() to obtain a Boolean value."
        )



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import ClassVar

from qilisdk.analog.hamiltonian import Hamiltonian, PauliOperator
from qilisdk.analog.schedule import Schedule
from qilisdk.core.qtensor import QTensor
from qilisdk.core.variables import RealNumber
from qilisdk.functionals.functional import PrimitiveFunctional
from qilisdk.functionals.time_evolution_result import TimeEvolutionResult
from qilisdk.yaml import yaml


@yaml.register_class
class TimeEvolution(PrimitiveFunctional[TimeEvolutionResult]):
    """
    Simulate the dynamics induced by a time-dependent Hamiltonian schedule.

    Example:
        .. code-block:: python

            from qilisdk.analog import Schedule, Hamiltonian, Z
            from qilisdk.core import ket
            from qilisdk.functionals.time_evolution import TimeEvolution

            h0 = Z(0)
            schedule = Schedule(T=10.0, hamiltonians={"h0": h0})
            functional = TimeEvolution(schedule, observables=[Z(0), X(0)], initial_state=ket(0))
    """

    result_type: ClassVar[type[TimeEvolutionResult]] = TimeEvolutionResult

    def __init__(
        self,
        schedule: Schedule,
        observables: list[PauliOperator | Hamiltonian],
        initial_state: QTensor,
        nshots: int = 1000,
        store_intermediate_results: bool = False,
    ) -> None:
        """
        Args:
            schedule (Schedule): Annealing or control schedule describing the Hamiltonian evolution.
            observables (list[PauliOperator | Hamiltonian]): Observables measured at the end of the evolution.
            initial_state (QTensor): Quantum state used as the simulation starting point.
            nshots (int, optional): Number of executions for statistical estimation. Defaults to 1000.
            store_intermediate_results (bool, optional): Keep intermediate states if produced by the backend. Defaults to False.
        """
        super().__init__()
        self.initial_state = initial_state
        self.schedule = schedule
        self.observables = observables
        self.nshots = nshots
        self.store_intermediate_results = store_intermediate_results

    @property
    def nparameters(self) -> int:
        """Return the number of schedule parameters."""
        return self.schedule.nparameters

    def get_parameters(self) -> dict[str, RealNumber]:
        """Return the schedule parameters and their current value."""
        return self.schedule.get_parameters()

    def set_parameters(self, parameters: dict[str, RealNumber]) -> None:
        """Update a subset of schedule parameters."""
        self.schedule.set_parameters(parameters)

    def get_parameter_names(self) -> list[str]:
        """Return order-stable parameter labels from the schedule."""
        return self.schedule.get_parameter_names()

    def get_parameter_values(self) -> list[RealNumber]:
        """Return parameter values in the order provided by ``get_parameter_names``."""
        return self.schedule.get_parameter_values()

    def set_parameter_values(self, values: list[float]) -> None:
        """Assign all schedule parameters according to ``get_parameter_names`` order."""
        self.schedule.set_parameter_values(values)

    def get_parameter_bounds(self) -> dict[str, tuple[float, float]]:
        """Return current bounds for schedule parameters."""
        return self.schedule.get_parameter_bounds()

    def set_parameter_bounds(self, ranges: dict[str, tuple[float, float]]) -> None:
        """Update bounds for selected schedule parameters."""
        self.schedule.set_parameter_bounds(ranges)



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from qilisdk.analog.schedule import Schedule
    from qilisdk.core.variables import Number

from qilisdk.utils.visualization.style import ScheduleStyle


class MatplotlibScheduleRenderer:
    """Render a Schedule using matplotlib, with theme support."""

    def __init__(self, schedule: Schedule, ax: plt.Axes | None = None, *, style: ScheduleStyle | None = None) -> None:
        self.schedule = schedule
        self.style = style or ScheduleStyle()
        self.ax = ax or self._make_axes(self.style.dpi, self.style)

    def plot(self, ax: plt.Axes | None = None) -> None:
        """
        Plot the schedule coefficients for each Hamiltonian over time.
        Args:
            ax (plt.Axes | None): The matplotlib axes to plot on. Default is None.
        """
        style = self.style
        theme = style.theme
        facecolor = theme.background
        title_color = theme.on_background
        label_color = theme.on_background
        legend_facecolor = theme.surface
        legend_edgecolor = theme.border
        tick_color = theme.on_background

        # Set axes and figure background to theme
        self.ax.set_facecolor(facecolor)
        if hasattr(ax, "figure"):
            self.ax.figure.set_facecolor(facecolor)
        plots: dict[str, list[Number]] = {}
        T = self.schedule.T
        dt = self.schedule.dt
        hamiltonians = self.schedule.hamiltonians
        times = [i * dt for i in range(int(T / dt))]
        for h in hamiltonians:
            plots[h] = []
        for _t in range(int(T / dt)):
            t = _t * dt
            for h in hamiltonians:
                plots[h].append(self.schedule.get_coefficient(t, h))

        # Generate gradient colors between primary and accent
        def hex_to_rgb(hex_color: str) -> tuple[int, ...]:
            hex_color = hex_color.lstrip("#")
            return tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))

        def rgb_to_hex(rgb: tuple[int, ...]) -> str:
            return "#{:02x}{:02x}{:02x}".format(*rgb)

        def gradient_colors(start_hex: str, end_hex: str, n: int) -> list[str]:
            start_rgb = hex_to_rgb(start_hex)
            end_rgb = hex_to_rgb(end_hex)
            colors = []
            for i in range(n):
                ratio = i / max(n - 1, 1)
                rgb = tuple(int(start_rgb[j] + (end_rgb[j] - start_rgb[j]) * ratio) for j in range(3))
                colors.append(rgb_to_hex(rgb))
            return colors

        n_hams = len(hamiltonians)
        grad_colors = gradient_colors(theme.primary, theme.accent, n_hams)

        for idx, h in enumerate(hamiltonians):
            line_style = style.line_styles.get(h, style.default_line_style)
            marker = style.marker
            # If no color specified, use gradient color
            if "color" not in line_style:
                color = grad_colors[idx]
                line_style = {**line_style, "color": color}
            self.ax.plot(times, plots[h], label=h, marker=marker, markersize=style.marker_size, **line_style)
        if style.grid:
            grid_style = dict(style.grid_style)
            if "color" not in grid_style:
                grid_style["color"] = theme.surface_muted
            self.ax.grid(**grid_style)
        leg = self.ax.legend(
            loc=style.legend_loc,
            fontsize=style.legend_fontsize,
            frameon=style.legend_frame,
            facecolor=legend_facecolor,
            edgecolor=legend_edgecolor,
        )
        # Set legend text color to match theme text color
        if leg:
            for text in leg.get_texts():
                text.set_color(title_color)
        self.ax.set_title(
            self.style.title or "Schedule Plot",
            fontsize=style.title_fontsize,
            color=title_color,
            fontweight=style.fontweight,
            family=style.fontfamily,
        )
        self.ax.set_xlabel(
            style.xlabel,
            fontsize=style.label_fontsize,
            color=label_color,
            fontweight=style.fontweight,
            family=style.fontfamily,
        )
        self.ax.set_ylabel(
            style.ylabel,
            fontsize=style.label_fontsize,
            color=label_color,
            fontweight=style.fontweight,
            family=style.fontfamily,
        )
        self.ax.tick_params(axis="x", labelsize=style.xtick_fontsize, colors=tick_color)
        self.ax.tick_params(axis="y", labelsize=style.ytick_fontsize, colors=tick_color)
        if style.tight_layout:
            plt.tight_layout()
        plt.draw()

    def save(self, filename: str) -> None:  # thin wrapper
        """Save current figure to disk.

        Args:
            filename: Path to save the figure (e.g., 'circuit.png').
        """

        self.ax.figure.savefig(filename, bbox_inches="tight")  # type: ignore[union-attr]

    def show(self) -> None:  # noqa: PLR6301
        """Show the current figure."""

        plt.show()

    @staticmethod
    def _make_axes(dpi: int, style: ScheduleStyle) -> plt.Axes:
        """
        Create a new figure and axes with the given DPI.

        Args:
            style: Optional style configuration (for DPI).

        Returns:
            A newly created Matplotlib Axes.
        """
        _, ax = plt.subplots(figsize=style.figsize, dpi=style.dpi, facecolor=style.theme.background)
        return ax



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Callable, Type, TypeVar

import numpy as np
import qutip_qip.operations as QutipGates
from loguru import logger
from qutip import Qobj, basis, mesolve, qeye, tensor
from qutip_qip.circuit import CircuitSimulator, QubitCircuit
from qutip_qip.operations.gateclass import SingleQubitGate, is_qutip5

from qilisdk.analog.hamiltonian import Hamiltonian, PauliI, PauliOperator
from qilisdk.backends.backend import Backend
from qilisdk.core.qtensor import QTensor, tensor_prod
from qilisdk.digital import RX, RY, RZ, SWAP, U1, U2, U3, Circuit, H, I, M, S, T, X, Y, Z
from qilisdk.digital.exceptions import UnsupportedGateError
from qilisdk.digital.gates import Adjoint, BasicGate, Controlled
from qilisdk.functionals.sampling_result import SamplingResult
from qilisdk.functionals.time_evolution_result import TimeEvolutionResult

if TYPE_CHECKING:
    from qilisdk.functionals.sampling import Sampling
    from qilisdk.functionals.time_evolution import TimeEvolution


TBasicGate = TypeVar("TBasicGate", bound=BasicGate)
BasicGateHandlersMapping = dict[Type[TBasicGate], Callable[[QubitCircuit, TBasicGate, int], None]]

TPauliOperator = TypeVar("TPauliOperator", bound=PauliOperator)
PauliOperatorHandlersMapping = dict[Type[TPauliOperator], Callable[[TPauliOperator], Qobj]]


class QutipI(SingleQubitGate):
    """
    Single-qubit I gate.

    Examples
    --------
    >>> from qutip_qip.operations import X
    >>> I(0).get_compact_qobj()  # doctest: +NORMALIZE_WHITESPACE
    Quantum object: dims=[[2], [2]], shape=(2, 2), type='oper', dtype=Dense, isherm=True
    Qobj data =
    [[1. 0.]
     [0. 1.]]
    """

    def __init__(self, targets, **kwargs) -> None:  # noqa: ANN001, ANN003
        super().__init__(targets=targets, **kwargs)
        self.name = "I"
        self.latex_str = r"I"

    def get_compact_qobj(self):  # noqa: ANN201, PLR6301
        return qeye(2) if not is_qutip5 else qeye(2, dtype="dense")


class QutipBackend(Backend):
    """
    Backend that runs both digital-circuit sampling and analog
    time-evolution experiments using the **QuTiP** simulation library.

    The backend is CPU-only and has no hardware dependencies, which makes it
    ideal for local development, CI pipelines, and educational notebooks.
    """

    def __init__(self, nsteps: int = 10_000) -> None:
        """Instantiate a new :class:`QutipBackend`.
        Args:
            nsteps (int): The maximum number of internal steps for the ODE solver."""
        self.nsteps = nsteps

        super().__init__()
        self._basic_gate_handlers: BasicGateHandlersMapping = {
            I: QutipBackend._handle_I,
            X: QutipBackend._handle_X,
            Y: QutipBackend._handle_Y,
            Z: QutipBackend._handle_Z,
            H: QutipBackend._handle_H,
            S: QutipBackend._handle_S,
            T: QutipBackend._handle_T,
            RX: QutipBackend._handle_RX,
            RY: QutipBackend._handle_RY,
            RZ: QutipBackend._handle_RZ,
            U1: QutipBackend._handle_U1,
            U2: QutipBackend._handle_U2,
            U3: QutipBackend._handle_U3,
            SWAP: QutipBackend._handle_SWAP,  # type: ignore[dict-item]
        }
        logger.success("QutipBackend initialised")

    def _execute_sampling(self, functional: Sampling) -> SamplingResult:
        """
        Execute a quantum circuit and return the measurement results.

        This method applies the selected simulation method, translates the circuit's gates into
        CUDA operations via their respective handlers, runs the simulation, and returns the result
        as a QutipDigitalResult.

        Args:
            functional (Sampling): The Sampling function to execute.

        Returns:
            DigitalResult: A result object containing the measurement samples and computed probabilities.

        """
        logger.info("Executing Sampling (shots={})", functional.nshots)
        qutip_circuit = self._get_qutip_circuit(functional.circuit)

        counts: Counter[str] = Counter()
        init_state = tensor(*[basis(2, 0) for _ in range(functional.circuit.nqubits)])

        measurements_set = set()
        for m in functional.circuit.gates:
            if isinstance(m, M):
                measurements_set.update(list(m.target_qubits))

        measurements = sorted(measurements_set)

        sim = CircuitSimulator(qutip_circuit)

        res = sim.run_statistics(init_state)  # runs the full circuit for one shot
        _bits = res.cbits  # classical measurement bits
        bits = []
        probs = res.probabilities

        if sum(probs) != 1:
            probs /= sum(probs)

        if len(measurements) > 0:
            for b in _bits:
                aux = []
                for i in measurements:
                    aux.append(b[i])
                bits.append(aux)
        else:
            bits = _bits

        bits_list = ["".join(map(str, cb)) for cb in bits]

        rng = np.random.default_rng()
        samples = rng.choice(bits_list, size=functional.nshots, p=probs)
        samples_py = map(str, samples)

        counts = Counter(samples_py)

        logger.success("Sampling finished; {} distinct bitstrings", len(counts))
        return SamplingResult(nshots=functional.nshots, samples=dict(counts))

    def _execute_time_evolution(self, functional: TimeEvolution) -> TimeEvolutionResult:
        """computes the time evolution under of an initial state under the given schedule.

        Args:
            functional (TimeEvolution): The TimeEvolution functional to execute.

        Returns:
            TimeEvolutionResult: The results of the evolution.

        Raises:
            ValueError: if the initial state provided is invalid.
        """
        logger.info("Executing TimeEvolution (T={}, dt={})", functional.schedule.T, functional.schedule.dt)
        tlist = np.linspace(0, functional.schedule.T, int(functional.schedule.T / functional.schedule.dt))

        qutip_hamiltonians = []
        for hamiltonian in functional.schedule.hamiltonians.values():
            qutip_hamiltonians.append(
                Qobj(
                    hamiltonian.to_matrix().toarray(), dims=[[2 for _ in range(hamiltonian.nqubits)] for _ in range(2)]
                )
            )

        H_t = [
            [
                qutip_hamiltonians[i],
                np.array([functional.schedule.get_coefficient(t, h) for t in tlist]),
            ]
            for i, h in enumerate(functional.schedule.hamiltonians)
        ]
        state_dim = []
        if functional.initial_state.is_density_matrix():
            state_dim = [[2 for _ in range(functional.initial_state.nqubits)] for _ in range(2)]
        elif functional.initial_state.is_bra():
            state_dim = [[1], [2 for _ in range(functional.initial_state.nqubits)]]
        elif functional.initial_state.is_ket():
            state_dim = [[2 for _ in range(functional.initial_state.nqubits)], [1]]
        else:
            logger.error("Invalid initial state provided")
            raise ValueError("invalid initial state provided.")

        qutip_init_state = Qobj(functional.initial_state.dense, dims=state_dim)

        qutip_obs: list[Qobj] = []

        identity = QTensor(PauliI(0).matrix)
        for obs in functional.observables:
            aux_obs = None
            if isinstance(obs, PauliOperator):
                for i in range(functional.schedule.nqubits):
                    if aux_obs is None:
                        aux_obs = identity if i != obs.qubit else QTensor(obs.matrix)
                    else:
                        aux_obs = (
                            tensor_prod([aux_obs, identity])
                            if i != obs.qubit
                            else tensor_prod([aux_obs, QTensor(obs.matrix)])
                        )
            elif isinstance(obs, Hamiltonian):
                aux_obs = QTensor(obs.to_matrix())
                if obs.nqubits < functional.schedule.nqubits:
                    for _ in range(functional.schedule.nqubits - obs.nqubits):
                        aux_obs = tensor_prod([aux_obs, identity])
            elif isinstance(obs, QTensor):
                aux_obs = obs
            else:
                logger.error("Unsupported observable type {}", obs.__class__.__name__)
                raise ValueError(f"unsupported observable type of {obs.__class__}")
            if aux_obs is not None:
                qutip_obs.append(
                    Qobj(aux_obs.dense, dims=[[2 for _ in range(functional.schedule.nqubits)] for _ in range(2)])
                )

        results = mesolve(
            H=H_t,
            e_ops=qutip_obs,
            rho0=qutip_init_state,
            tlist=tlist,
            options={
                "store_states": functional.store_intermediate_results,
                "store_final_state": True,
                "nsteps": self.nsteps,
            },
        )

        logger.success("TimeEvolution finished")
        return TimeEvolutionResult(
            final_expected_values=np.array([results.expect[i][-1] for i in range(len(qutip_obs))]),
            expected_values=(
                np.array(
                    [
                        [results.expect[val][i] for val in range(len(results.expect))]
                        for i in range(len(results.expect[0]))
                    ]
                )
                if len(results.expect) > 0 and functional.store_intermediate_results
                else None
            ),
            final_state=(QTensor(results.final_state.full()) if results.final_state is not None else None),
            intermediate_states=(
                [QTensor(state.full()) for state in results.states]
                if len(results.states) > 1 and functional.store_intermediate_results
                else None
            ),
        )

    def _get_qutip_circuit(self, circuit: Circuit) -> QubitCircuit:
        """_summary_

        Args:
            circuit (Circuit): the qiliSDK circuit to be translated to qutip.

        Raises:
            UnsupportedGateError: If the circuit contains a gate for which no handler is registered.

        Returns:
            QubitCircuit: the translated qutip circuit.
        """
        qutip_circuit = QubitCircuit(
            circuit.nqubits, num_cbits=circuit.nqubits, input_states=[0 for _ in range(circuit.nqubits)]
        )

        for gate in circuit.gates:
            if isinstance(gate, Controlled):
                self._handle_controlled(qutip_circuit, gate)
            elif isinstance(gate, Adjoint):
                self._handle_adjoint(qutip_circuit, gate)
            elif isinstance(gate, M):
                self._handle_M(qutip_circuit, gate)
            else:
                handler = self._basic_gate_handlers.get(type(gate), None)
                if handler is None:
                    logger.error("Unsupported gate {}", type(gate).__name__)
                    raise UnsupportedGateError(f"Unsupported gate {type(gate).__name__}")
                handler(qutip_circuit, gate, *(qubit for qubit in gate.target_qubits))

        no_measurement = True

        for g in circuit.gates:
            if isinstance(g, M):
                no_measurement = False

        if no_measurement:
            for i in range(circuit.nqubits):
                qutip_circuit.add_measurement(f"M{i}", targets=i, classical_store=i)
        return qutip_circuit

    def _handle_controlled(self, circuit: QubitCircuit, gate: Controlled) -> None:  # noqa: PLR6301
        """
        Handle a controlled gate operation.

        This method processes a controlled gate by creating a temporary kernel for the basic gate,
        applying its handler, and then integrating it into the main kernel as a controlled operation.

        Raises:
            UnsupportedGateError: If the number of control qubits is not equal to one or if the basic gate is unsupported.
        """
        if len(gate.control_qubits) != 1:
            logger.error("Controlled gate with {} control qubits not supported", len(gate.control_qubits))
            raise UnsupportedGateError

        def qutip_controlled_gate() -> Qobj:
            return QutipGates.controlled_gate(Qobj(gate.basic_gate.matrix), controls=0, targets=1)

        if gate.name == "CNOT":
            circuit.add_gate("CNOT", targets=[*gate.target_qubits], controls=[*gate.control_qubits])
        else:
            gate_name = "Controlled_" + gate.name
            if gate_name not in circuit.user_gates:
                circuit.user_gates[gate_name] = qutip_controlled_gate
            circuit.add_gate(gate_name, targets=[*gate.control_qubits, *gate.target_qubits])

    def _handle_adjoint(self, circuit: QubitCircuit, gate: Adjoint) -> None:  # noqa: PLR6301
        """
        Handle an adjoint (inverse) gate operation.

        This method creates a temporary kernel for the basic gate wrapped by the adjoint,
        applies the corresponding handler, and then integrates it into the main kernel as an adjoint operation.
        """

        def qutip_adjoined_gate() -> Qobj:
            return Qobj(gate.matrix)

        gate_name = "Adjoint_" + gate.name
        if gate_name not in circuit.user_gates:
            circuit.user_gates[gate_name] = qutip_adjoined_gate
        circuit.add_gate(gate_name, targets=[*gate.target_qubits])

    @staticmethod
    def _handle_M(qutip_circuit: QubitCircuit, gate: M) -> None:
        """
        Handle a measurement gate.

        Depending on whether the measurement targets all qubits or a subset,
        this method applies measurement operations accordingly.
        """
        for i in gate.target_qubits:
            qutip_circuit.add_measurement(f"M{i}", targets=[i], classical_store=i)

    @staticmethod
    def _handle_I(circuit: QubitCircuit, gate: I, qubit: int) -> None:
        """Handle an X gate operation."""
        circuit.add_gate(QutipI(targets=qubit))

    @staticmethod
    def _handle_X(circuit: QubitCircuit, gate: X, qubit: int) -> None:
        """Handle an X gate operation."""
        circuit.add_gate(QutipGates.X(targets=qubit))

    @staticmethod
    def _handle_Y(circuit: QubitCircuit, gate: Y, qubit: int) -> None:
        """Handle an Y gate operation."""
        circuit.add_gate(QutipGates.Y(targets=qubit))

    @staticmethod
    def _handle_Z(circuit: QubitCircuit, gate: Z, qubit: int) -> None:
        """Handle an Z gate operation."""
        circuit.add_gate(QutipGates.Z(targets=qubit))

    @staticmethod
    def _handle_H(circuit: QubitCircuit, gate: H, qubit: int) -> None:
        """Handle an H gate operation."""
        circuit.add_gate(QutipGates.H(targets=qubit))

    @staticmethod
    def _handle_S(circuit: QubitCircuit, gate: S, qubit: int) -> None:
        """Handle an S gate operation."""
        circuit.add_gate(QutipGates.S(targets=qubit))

    @staticmethod
    def _handle_T(circuit: QubitCircuit, gate: T, qubit: int) -> None:
        """Handle an T gate operation."""
        circuit.add_gate(QutipGates.T(targets=qubit))

    @staticmethod
    def _handle_RX(circuit: QubitCircuit, gate: RX, qubit: int) -> None:
        """Handle an RX gate operation."""
        circuit.add_gate(QutipGates.RX(targets=[qubit], arg_value=gate.get_parameter_values()[0]))

    @staticmethod
    def _handle_RY(circuit: QubitCircuit, gate: RY, qubit: int) -> None:
        """Handle an RY gate operation."""
        circuit.add_gate(QutipGates.RY(targets=[qubit], arg_value=gate.get_parameter_values()[0]))

    @staticmethod
    def _handle_RZ(circuit: QubitCircuit, gate: RZ, qubit: int) -> None:
        """Handle an RZ gate operation."""
        circuit.add_gate(QutipGates.RZ(targets=[qubit], arg_value=gate.get_parameter_values()[0]))

    @staticmethod
    def _qutip_U1(phi: float) -> Qobj:
        mat = np.array([[1, 0], [0, np.exp(1j * phi)]], dtype=complex)
        return Qobj(mat, dims=[[2], [2]])

    @staticmethod
    def _handle_U1(circuit: QubitCircuit, gate: U1, qubit: int) -> None:
        """Handle an U1 gate operation."""
        U1_label = "U1"

        if U1_label not in circuit.user_gates:
            circuit.user_gates[U1_label] = QutipBackend._qutip_U1
        circuit.add_gate(U1_label, targets=qubit, arg_value=gate.phi)

    @staticmethod
    def _qutip_U2(angles: list[float]) -> Qobj:
        phi = angles[0]
        gamma = angles[1]
        mat = (1 / np.sqrt(2)) * np.array(
            [
                [1, -np.exp(1j * gamma)],
                [np.exp(1j * phi), np.exp(1j * (phi + gamma))],
            ],
            dtype=complex,
        )
        return Qobj(mat, dims=[[2], [2]])

    @staticmethod
    def _handle_U2(circuit: QubitCircuit, gate: U2, qubit: int) -> None:
        """Handle an U2 gate operation."""
        U2_label = "U2"

        if U2_label not in circuit.user_gates:
            circuit.user_gates[U2_label] = QutipBackend._qutip_U2
        circuit.add_gate(U2_label, targets=qubit, arg_value=[gate.phi, gate.gamma])

    @staticmethod
    def _qutip_U3(angles: list[float]) -> Qobj:
        phi = angles[0]
        gamma = angles[1]
        theta = angles[2]
        mat = np.array(
            [
                [np.cos(theta / 2), -np.exp(1j * gamma) * np.sin(theta / 2)],
                [np.exp(1j * phi) * np.sin(theta / 2), np.exp(1j * (phi + gamma)) * np.cos(theta / 2)],
            ],
            dtype=complex,
        )
        return Qobj(mat, dims=[[2], [2]])

    @staticmethod
    def _handle_U3(circuit: QubitCircuit, gate: U3, qubit: int) -> None:
        """Handle an U3 gate operation."""
        U3_label = "U3"

        if U3_label not in circuit.user_gates:
            circuit.user_gates[U3_label] = QutipBackend._qutip_U3
        circuit.add_gate(U3_label, targets=qubit, arg_value=[gate.phi, gate.gamma, gate.theta])

    @staticmethod
    def _handle_SWAP(circuit: QubitCircuit, gate: SWAP, qubit_0: int, qubit_1: int) -> None:
        """Handle a SWAP gate operation."""
        circuit.add_gate(QutipGates.SWAP(targets=[qubit_0, qubit_1]))



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from qilisdk.core.parameterizable import Parameterizable
from qilisdk.core.variables import Parameter, RealNumber
from qilisdk.utils.visualization import CircuitStyle
from qilisdk.yaml import yaml

from .exceptions import ParametersNotEqualError, QubitOutOfRangeError
from .gates import Gate


@yaml.register_class
class Circuit(Parameterizable):
    def __init__(self, nqubits: int) -> None:
        """
        Initialize a Circuit instance with a specified number of qubits.

        Args:
            nqubits (int): The number of qubits in the circuit.
        """
        self._nqubits: int = nqubits
        self._gates: list[Gate] = []
        self._init_state: np.ndarray = np.zeros(nqubits)
        self._parameters: dict[str, Parameter] = {}

    @property
    def nqubits(self) -> int:
        """
        Retrieve the number of qubits in the circuit.

        Returns:
            int: The total number of qubits.
        """
        return self._nqubits

    @property
    def nparameters(self) -> int:
        """
        Retrieve the total number of parameters required by all parameterized gates in the circuit.

        Returns:
            int: The total count of parameters from all parameterized gates.
        """
        return len(self._parameters)

    @property
    def gates(self) -> list[Gate]:
        """
        Retrieve the list of gates in the circuit.

        Returns:
            list[Gate]: A list of gates that have been added to the circuit.
        """
        return self._gates

    def get_parameter_values(self) -> list[float]:
        """
        Retrieve the parameter values from all parameterized gates in the circuit.

        Returns:
            list[float]: A list of parameter values from each parameterized gate.
        """
        return [param.value for param in self._parameters.values()]

    def get_parameter_names(self) -> list[str]:
        """
        Retrieve the parameter values from all parameterized gates in the circuit.

        Returns:
            list[float]: A list of parameter values from each parameterized gate.
        """
        return list(self._parameters.keys())

    def get_parameters(self) -> dict[str, float]:
        """
        Retrieve the parameter names and values from all parameterized gates in the circuit.

        Returns:
            dict[str, float]: A dictionary of the parameters with their current values.
        """
        return {label: param.value for label, param in self._parameters.items()}

    def set_parameter_values(self, values: list[float]) -> None:
        """
        Set new parameter values for all parameterized gates in the circuit.

        Args:
            values (list[float]): A list containing new parameter values to assign to the parameterized gates.

        Raises:
            ParametersNotEqualError: If the number of provided values does not match the expected number of parameters.
        """
        if len(values) != self.nparameters:
            raise ParametersNotEqualError
        for i, parameter in enumerate(self._parameters.values()):
            parameter.set_value(values[i])

    def set_parameters(self, parameter_dict: dict[str, RealNumber]) -> None:
        """Set the parameter values by their label. No need to provide the full list of parameters.

        Args:
            parameter_dict (dict[str, RealNumber]): A dictionary with the labels of the parameters to be modified and their new value.

        Raises:
            ValueError: if the label provided doesn't correspond to a parameter defined in this circuit.
        """
        for label, param in parameter_dict.items():
            if label not in self._parameters:
                raise ValueError(f"Parameter {label} is not defined in this circuit.")
            self._parameters[label].set_value(param)

    def get_parameter_bounds(self) -> dict[str, tuple[float, float]]:
        return {k: v.bounds for k, v in self._parameters.items()}

    def set_parameter_bounds(self, ranges: dict[str, tuple[float, float]]) -> None:
        for label, bound in ranges.items():
            if label not in self._parameters:
                raise ValueError(
                    f"The provided parameter label {label} is not defined in the list of parameters in this object."
                )
            self._parameters[label].set_bounds(bound[0], bound[1])

    def add(self, gate: Gate) -> None:
        """
        Add a quantum gate to the circuit.

        Args:
            gate (Gate): The quantum gate to add to the circuit.

        Raises:
            QubitOutOfRangeError: If any qubit index used by the gate is not within the circuit's qubit range.
        """
        if any(qubit >= self.nqubits for qubit in gate.qubits):
            raise QubitOutOfRangeError
        if gate.is_parameterized:
            param_base_label = f"{gate.name}({','.join(map(str, gate.qubits))})"
            for label, parameter in gate.parameters.items():
                if label == parameter.label:
                    parameter_label = param_base_label + f"_{label}_{len(self._parameters)}"
                else:
                    parameter_label = parameter.label
                self._parameters[parameter_label] = gate.parameters[label]
        self._gates.append(gate)

    def draw(self, style: CircuitStyle = CircuitStyle(), filepath: str | None = None) -> None:
        """
        Render this circuit with Matplotlib and optionally save it to a file.

        The circuit is rendered using the provided style configuration. If ``filepath`` is
        given, the resulting figure is saved to disk (the output format is inferred
        from the file extension, e.g. ``.png``, ``.pdf``, ``.svg``).

        Args:
            style (CircuitStyle): Visual style configuration applied to the plot.
                If not provided, the default :class:`CircuitStyle` is used.
            filepath (str | None): Destination file path for the rendered figure.
                If ``None``, the figure is not saved.
        """
        from qilisdk.utils.visualization.circuit_renderers import MatplotlibCircuitRenderer  # noqa: PLC0415

        renderer = MatplotlibCircuitRenderer(self, style=style)
        renderer.plot()
        if filepath:
            renderer.save(filepath)



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import heapq
import operator
from pprint import pformat

from qilisdk.functionals.functional_result import FunctionalResult
from qilisdk.yaml import yaml


@yaml.register_class
class SamplingResult(FunctionalResult):
    """Store shot counts and derived probabilities for a sampling experiment."""

    def __init__(self, nshots: int, samples: dict[str, int]) -> None:
        """
        Args:
            nshots (int): Total number of circuit evaluations.
            samples (dict[str, int]): Mapping from bitstring to observed counts.

        Raises:
            ValueError: If ``samples`` is empty or contains bitstrings with different length.
        """
        self._nshots = nshots
        self._samples = samples

        # Ensure samples is not empty and is correct.
        if not samples:
            raise ValueError("The samples dictionary is empty.")
        bitstrings = list(samples.keys())
        nqubits = len(bitstrings[0])
        if not all(len(bitstring) == nqubits for bitstring in bitstrings):
            raise ValueError("Not all bitstring keys have the same length.")

        # Assign nqubits to attribute
        self._nqubits = nqubits

        # Calculate probabilities
        self._probabilities = {
            bitstring: counts / self._nshots if self._nshots > 0 else 0.0 for bitstring, counts in self._samples.items()
        }

    @property
    def nshots(self) -> int:
        """Total number of repetitions used to gather samples."""
        return self._nshots

    @property
    def nqubits(self) -> int:
        """Number of qubits inferred from the sample bitstrings."""
        return self._nqubits

    @property
    def samples(self) -> dict[str, int]:
        """Return a copy of the raw sample counts."""
        return dict(self._samples)

    @property
    def probabilities(self) -> dict[str, float]:
        """Return a copy of the estimated probability distribution."""
        return dict(self._probabilities)

    def get_probability(self, bitstring: str) -> float:
        """Return the probability associated with ``bitstring`` (0.0 if unseen)."""
        return self._probabilities.get(bitstring, 0.0)

    def get_probabilities(self, n: int | None = None) -> list[tuple[str, float]]:
        """
        Args:
            n (int | None): Maximum number of items to return. Defaults to all outcomes.

        Returns:
            list[tuple[str, float]]: the ``n`` most probable bitstrings in descending probability order.
        """
        if n is None:
            n = len(self._probabilities)
        return heapq.nlargest(n, self._probabilities.items(), key=operator.itemgetter(1))

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(\n  nshots={self.nshots},\n  samples={pformat(self.samples)}\n)"



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Callable, TypeVar, cast, overload

from qilisdk.functionals.functional_result import FunctionalResult
from qilisdk.functionals.sampling import Sampling
from qilisdk.functionals.time_evolution import TimeEvolution
from qilisdk.functionals.variational_program import VariationalProgram
from qilisdk.functionals.variational_program_result import VariationalProgramResult

if TYPE_CHECKING:
    from qilisdk.functionals.functional import Functional, PrimitiveFunctional
    from qilisdk.functionals.sampling_result import SamplingResult
    from qilisdk.functionals.time_evolution_result import TimeEvolutionResult

TResult = TypeVar("TResult", bound=FunctionalResult)


class Backend(ABC):
    def __init__(self) -> None:
        self._handlers: dict[type[Functional], Callable[[Functional], FunctionalResult]] = {
            Sampling: lambda f: self._execute_sampling(cast("Sampling", f)),
            TimeEvolution: lambda f: self._execute_time_evolution(cast("TimeEvolution", f)),
            VariationalProgram: lambda f: self._execute_variational_program(cast("VariationalProgram", f)),
        }

    @overload
    def execute(self, functional: Sampling) -> SamplingResult: ...

    @overload
    def execute(self, functional: TimeEvolution) -> TimeEvolutionResult: ...

    @overload
    def execute(self, functional: VariationalProgram[Sampling]) -> VariationalProgramResult[SamplingResult]: ...

    @overload
    def execute(
        self, functional: VariationalProgram[TimeEvolution]
    ) -> VariationalProgramResult[TimeEvolutionResult]: ...

    @overload
    def execute(self, functional: PrimitiveFunctional[TResult]) -> TResult: ...

    def execute(self, functional: Functional) -> FunctionalResult:
        try:
            handler = self._handlers[type(functional)]
        except KeyError as exc:
            raise NotImplementedError(
                f"{type(self).__qualname__} does not support {type(functional).__qualname__}"
            ) from exc

        return handler(functional)

    def _execute_sampling(self, functional: Sampling) -> SamplingResult:
        raise NotImplementedError(f"{type(self).__qualname__} has no Sampling implementation")

    def _execute_time_evolution(self, functional: TimeEvolution) -> TimeEvolutionResult:
        raise NotImplementedError(f"{type(self).__qualname__} has no TimeEvolution implementation")

    def _execute_variational_program(
        self, functional: VariationalProgram[PrimitiveFunctional[TResult]]
    ) -> VariationalProgramResult[TResult]:
        """Optimize a Parameterized Program (:class:`~qilisdk.functionals.variational_program.VariationalProgram`)
            and returns the optimal parameters and results.

        Args:
            functional (VariationalProgram): The variational program to be optimized.

        Returns:
            ParameterizedProgramResults: The final optimizer and functional results.

        Raises:
            ValueError: If the functional is not parameterized.
        """

        def evaluate_sample(parameters: list[float]) -> float:
            param_names = functional.functional.get_parameter_names()
            functional.functional.set_parameters({param_names[i]: param for i, param in enumerate(parameters)})
            results = self.execute(functional.functional)
            final_results = functional.cost_function.compute_cost(results)
            if isinstance(final_results, float):
                return final_results
            if isinstance(final_results, complex) and final_results.imag == 0:
                return final_results.real
            raise ValueError(f"Unsupported result type {type(final_results)}.")

        if len(functional.functional.get_parameters()) == 0:
            raise ValueError("Functional provided is not parameterized.")

        optimizer_result = functional.optimizer.optimize(
            cost_function=evaluate_sample,
            init_parameters=list(functional.functional.get_parameters().values()),
            bounds=list(functional.functional.get_parameter_bounds().values()),
            store_intermediate_results=functional.store_intermediate_results,
        )

        param_names = functional.functional.get_parameter_names()
        functional.functional.set_parameters(
            {param_names[i]: param for i, param in enumerate(optimizer_result.optimal_parameters)}
        )
        optimal_results: TResult = cast("TResult", self.execute(functional.functional))

        return VariationalProgramResult(optimizer_result=optimizer_result, result=optimal_results)



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""config.py"""

from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger
from pydantic import BaseModel
from pydantic_settings import BaseSettings
from ruamel.yaml import YAML

from qilisdk.settings import get_settings

if TYPE_CHECKING:
    from types import FrameType


class SinkConfig(BaseModel):
    """
    Configuration for a single Loguru sink.
    """

    sink: str | Path
    level: str = "INFO"
    format: str | None = None
    filter: str | dict[str, str] | None = None
    colorize: bool = False
    enqueue: bool = False
    rotation: str | None = None
    serialize: bool = False


class InterceptLibraryConfig(BaseModel):
    """
    Configuration for intercepting a stdlib logging library.
    """

    name: str
    level: str = "ERROR"


class LoggingSettings(BaseSettings):
    """
    Pydantic settings for Loguru configuration loaded from YAML or JSON.
    """

    sinks: list[SinkConfig] = []
    intercept_libraries: list[InterceptLibraryConfig] = []

    @classmethod
    def load(cls, path: str | Path) -> LoggingSettings:
        path = Path(path)
        yaml = YAML(typ="safe")
        data = yaml.load(path)
        return cls(**data)


class InterceptHandler(logging.Handler):
    """
    Redirect stdlib 'logging' records to Loguru, optionally filtering by name_prefix.
    """

    def emit(self, record: logging.LogRecord) -> None:  # noqa: PLR6301
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno  # type: ignore[assignment]

        frame: FrameType | None = logging.currentframe()
        depth = 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def configure_logging() -> None:
    """
    Load settings (path overridden by environment variable) and configure Loguru + stdlib logging intercept.
    """
    # Determine config path
    config_path = Path(get_settings().logging_config_path).expanduser()
    if not config_path.is_absolute():
        config_path = (Path.cwd() / config_path).resolve()

    settings = LoggingSettings.load(config_path)

    # 1) Remove all pre-configured Loguru handlers
    logger.remove()

    # 2) Add configured sinks
    for sink_conf in settings.sinks:
        params = sink_conf.model_dump()
        sink_target = params.pop("sink")

        # Resolve stderr/stdout
        if isinstance(sink_target, str) and sink_target.lower() == "stderr":
            sink_target = sys.stderr
        elif isinstance(sink_target, str) and sink_target.lower() == "stdout":
            sink_target = sys.stdout

        clean_params = {k: v for k, v in params.items() if v is not None}

        logger.add(sink_target, **clean_params)

    # 3) Quiet down noisy libraries before intercepting
    for intercept_library in settings.intercept_libraries:
        logging.getLogger(intercept_library.name).setLevel(intercept_library.level)

    # 4) Intercept stdlib -> Loguru
    handler = InterceptHandler()
    logging.root.handlers = [handler]
    logging.root.setLevel(logging.NOTSET)

    for logger_name in list(logging.root.manager.loggerDict.keys()):
        logging.getLogger(logger_name).handlers = []
        logging.getLogger(logger_name).propagate = True



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Callable, TypeVar, cast, overload

from qilisdk.functionals.functional_result import FunctionalResult
from qilisdk.functionals.sampling_result import SamplingResult
from qilisdk.functionals.time_evolution_result import TimeEvolutionResult

if TYPE_CHECKING:
    from qilisdk.core.variables import Number

TResult = TypeVar("TResult", bound=FunctionalResult)


class CostFunction(ABC):
    """
    Base class that maps functional results into scalar costs.
    """

    def __init__(self) -> None:
        self._handlers: dict[type[FunctionalResult], Callable[[FunctionalResult], Number]] = {
            SamplingResult: lambda f: self._compute_cost_sampling(cast("SamplingResult", f)),
            TimeEvolutionResult: lambda f: self._compute_cost_time_evolution(cast("TimeEvolutionResult", f)),
        }

    @overload
    def compute_cost(self, results: SamplingResult) -> Number: ...

    @overload
    def compute_cost(self, results: TimeEvolutionResult) -> Number: ...

    @overload
    def compute_cost(self, results: FunctionalResult) -> Number: ...

    def compute_cost(self, results: TResult) -> Number:
        """
        Dispatch to the appropriate cost implementation based on the result type.

        Args:
            results (FunctionalResult): Output of a functional execution.

        Returns:
            Number: Scalar cost extracted from the results.

        Raises:
            NotImplementedError: If the concrete cost function does not support the given result type.
        """
        try:
            handler = self._handlers[type(results)]
        except KeyError as exc:
            raise NotImplementedError(
                f"{type(self).__qualname__} does not support {type(results).__qualname__}"
            ) from exc

        return handler(results)

    def _compute_cost_sampling(self, results: SamplingResult) -> Number:
        """Compute the cost associated with a :class:`SamplingResult`."""
        raise NotImplementedError

    def _compute_cost_time_evolution(self, results: TimeEvolutionResult) -> Number:
        """Compute the cost associated with a :class:`TimeEvolutionResult`."""
        raise NotImplementedError



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Literal, Optional

import matplotlib.font_manager as fm
from pydantic import BaseModel, Field

from .themes import Theme, light

_DEFAULT_FONT_PATH = Path(__file__).parent / "PlusJakartaSans-SemiBold.ttf"


class Style(BaseModel):
    # --- FontProperties-mapped fields (mirror matplotlib.font_manager.FontProperties) ---
    # If `fontfname` exists, it takes precedence and loads the exact TTF.
    theme: Theme = Field(default=light, description="Colour theme.")
    fontfamily: str | list[str] | None = Field(
        default=None, description="Font family name(s), e.g. 'Outfit' or ['Outfit', 'DejaVu Sans']."
    )
    fontstyle: Literal["normal", "italic", "oblique"] = Field(
        default="normal", description="Font style: 'normal', 'italic', or 'oblique'."
    )
    fontvariant: Literal["normal", "small-caps"] = Field(
        default="normal", description="Font variant: typically 'normal' or 'small-caps'."
    )
    fontweight: str | int = Field(
        default="normal", description="Font weight: 'normal', 'bold', 'light', or numeric (100-900)."
    )
    fontstretch: str | int = Field(
        default="normal", description="Width/condensation: 'ultra-condensed'..'ultra-expanded' or numeric."
    )
    fontsize: float | str = Field(
        default=10, description="Font size in pt or keywords like 'small', 'medium', 'large'."
    )
    fontfname: str | None = Field(
        default=str(_DEFAULT_FONT_PATH), description="Absolute path to the TTF/OTF file. If present, overrides family."
    )
    math_fontfamily: str | None = Field(default=None, description="Math text family, e.g. 'dejavusans', 'cm', or None.")
    dpi: int = Field(default=150, description="Figure DPI.")
    title: str | None = Field(default=None, description="Figure title.")

    @property
    def font(self) -> fm.FontProperties:
        """
        Construct a Matplotlib FontProperties from the configured fields.
        If `fontfname` points to a real file, it is used (and overrides family).
        """
        return fm.FontProperties(
            family=self.fontfamily,
            style=self.fontstyle,
            variant=self.fontvariant,
            weight=self.fontweight,
            stretch=self.fontstretch,
            size=self.fontsize,
            fname=self.fontfname,
            math_fontfamily=self.math_fontfamily,
        )


class CircuitStyle(Style):
    """All visual parameters controlling the appearance of a circuit plot."""

    end_wire_ext: int = Field(default=2, description="Extra space after last layer.")
    padding: float = Field(default=0.3, description="Padding around drawing (inches).")
    gate_margin: float = Field(default=0.15, description="Left/right margin per gate.")
    wire_sep: float = Field(default=0.5, description="Vertical separation of wires.")
    layer_sep: float = Field(default=0.5, description="Horizontal separation of layers.")
    gate_pad: float = Field(default=0.05, description="Padding around gate text.")
    label_pad: float = Field(default=0.1, description="Padding before wire label.")
    bulge: str = Field(default="round", description="Box-style for gate rectangles.")
    align_layer: bool = Field(default=True, description="Align layers across wires.")

    wire_label: list[Any] | None = Field(default=None, description="Custom wire labels.")
    start_pad: float = Field(
        default=0.1, description="Minimum spacing (inches) before the first layer so wire labels fit."
    )
    min_gate_h: float = Field(default=0.2, description="Minimum gate box height (inches).")
    min_gate_w: float = Field(default=0.2, description="Minimum gate box width (inches).")
    connector_r: float = Field(
        default=0.01, description="Radius (inches) of small connector dots on multi-target gates."
    )
    target_r: float = Field(default=0.12, description="Radius (inches) of ⊕ target circle and SWAP half-width.")
    control_r: float = Field(default=0.05, description="Radius (inches) of a filled control dot.")

    layout: Literal["normal", "compact"] = Field(
        default="normal",
        description="If 'compact' minimizes the layers to highlight circuit depth, if 'normal' conserves the order of the circuit",
    )


class ScheduleStyle(Style):
    """
    Customization options for matplotlib schedule plots, with theme support.
    """

    # Figure and axes
    figsize: Optional[tuple] = Field(default=(8, 5), description="Figure size in inches (width, height).")
    grid: bool = Field(default=True, description="Whether to show grid lines on the plot.")
    grid_style: dict[str, Any] = Field(
        default_factory=lambda: {"linestyle": "--", "color": "#e0e0e0", "alpha": 0.7},
        description="Style dictionary for grid lines (linestyle, color, alpha, etc.).",
    )

    # Title and labels
    title_fontsize: int = Field(default=16, description="Font size for the plot title.")
    xlabel: str = Field(default="time (dt)", description="Label for the x-axis.")
    ylabel: str = Field(default="coefficient value", description="Label for the y-axis.")
    label_fontsize: int = Field(default=14, description="Font size for axis labels.")

    # Legend
    legend_loc: str = Field(
        default="best", description="Location of the legend (matplotlib string, e.g. 'best', 'upper right')."
    )
    legend_fontsize: int = Field(default=12, description="Font size for legend text.")
    legend_frame: bool = Field(default=True, description="Whether to draw a frame around the legend.")

    # Line style
    line_styles: dict[str, dict[str, Any]] = Field(
        default_factory=dict,
        description="Custom line style dictionary for each Hamiltonian (e.g. {label: {color, linestyle, linewidth}}).",
    )
    default_line_style: dict[str, Any] = Field(
        default_factory=lambda: {"linestyle": "-", "linewidth": 2},
        description="Default line style for Hamiltonians not in line_styles.",
    )

    # Marker style
    marker: Optional[str] = Field(
        default=None, description="Matplotlib marker style for data points (e.g. 'o', 's', None for no marker)."
    )
    marker_size: int = Field(default=6, description="Size of markers if used.")

    # Ticks
    xtick_fontsize: int = Field(default=12, description="Font size for x-axis tick labels.")
    ytick_fontsize: int = Field(default=12, description="Font size for y-axis tick labels.")
    tick_color: Optional[str] = Field(
        default=None, description="Color for tick labels (None uses theme.on_background)."
    )

    # Misc
    tight_layout: bool = Field(default=True, description="Whether to use matplotlib's tight_layout for figure spacing.")



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import ABC
from typing import ClassVar, Generic, TypeVar

from qilisdk.core.parameterizable import Parameterizable
from qilisdk.functionals.functional_result import FunctionalResult

TResult_co = TypeVar("TResult_co", bound=FunctionalResult, covariant=True)


class Functional(ABC):
    """
    Abstract interface for executable routines that return a :class:`FunctionalResult`.

    Subclasses detail the concrete `result_type` they generate.
    """

    result_type: ClassVar[type[FunctionalResult]]
    """Concrete :class:`~qilisdk.functionals.functional_result.FunctionalResult` subclass returned."""


class PrimitiveFunctional(Parameterizable, Functional, ABC, Generic[TResult_co]):
    """
    Base class for functionals backed by a :class:`~qilisdk.core.parameterizable.Parameterizable` object.
    """



# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from qilisdk.analog.hamiltonian import Hamiltonian, PauliOperator
from qilisdk.core.qtensor import QTensor, expect_val, ket, tensor_prod
from qilisdk.cost_functions.cost_function import CostFunction

if TYPE_CHECKING:
    from qilisdk.core.variables import Number
    from qilisdk.functionals.sampling_result import SamplingResult
    from qilisdk.functionals.time_evolution_result import TimeEvolutionResult


class ObservableCostFunction(CostFunction):
    """
    Compute costs by taking expectation values of observables.

    Example:
        .. code-block:: python

            from qilisdk.analog.hamiltonian import Z
            from qilisdk.cost_functions import ObservableCostFunction

            cost_fn = ObservableCostFunction(Z(0))
    """

    def __init__(self, observable: QTensor | Hamiltonian | PauliOperator) -> None:
        """
        Args:
            observable (QTensor | Hamiltonian | PauliOperator): Quantum observable whose expectation value defines the cost.

        Raises:
            ValueError: If the provided observable type is unsupported.
        """
        super().__init__()
        if isinstance(observable, QTensor):
            self._observable = observable
        elif isinstance(observable, Hamiltonian):
            self._observable = QTensor(observable.to_matrix())
        elif isinstance(observable, PauliOperator):
            self._observable = QTensor(observable.matrix)
        else:
            raise ValueError(
                f"Observable needs to be of type QTensor, Hamiltonian, or PauliOperator but {type(observable)} was provided"
            )

    @property
    def observable(self) -> QTensor:
        """Return the observable in ``QTensor`` form."""
        return self._observable

    def _compute_cost_time_evolution(self, results: TimeEvolutionResult) -> Number:
        """
        Returns:
            Number: the expectation value of ``observable`` given a final quantum state.

        Raises:
            ValueError: If the final state is not provided in the results.

        """
        if results.final_state is None:
            raise ValueError(
                "can't compute cost using Observables from time evolution results when the state is not provided."
            )
        total_cost = complex(np.real_if_close(expect_val(self._observable, results.final_state)))
        if total_cost.imag == 0:
            return total_cost.real
        return total_cost

    def _compute_cost_sampling(self, results: SamplingResult) -> Number:
        """
        Estimate the observable expectation value using measured bitstring probabilities.

        Returns:
            Number: the expectation value of ``observable`` estimated from sampled bitstrings.

        Raises:
            ValueError: If the number of qubits in the observable does not match the sample size
        """
        total_cost = complex(0.0)
        nqubits = self._observable.nqubits
        for sample, prob in results.get_probabilities():
            state = tensor_prod([ket(int(i)) for i in sample])
            if nqubits != state.nqubits:
                raise ValueError(
                    f"The samples provided have {state.nqubits} qubits but the observable has {nqubits} qubits"
                )
            evaluate_results = complex(np.real_if_close(expect_val(self._observable, state)))
            total_cost += evaluate_results * prob

        if total_cost.imag == 0:
            return total_cost.real
        return total_cost
